import asyncio

import pytest

from storey.queue import SimpleAsyncQueue


async def async_test_simple_async_queue():
    q = SimpleAsyncQueue(2)

    with pytest.raises(TimeoutError):
        await q.get(0)

    get_task = asyncio.create_task(q.get(1))
    await q.put("x")
    assert await get_task == "x"

    await q.put("x")
    await q.put("y")
    put_task = asyncio.create_task(q.put("z"))
    assert await q.get() == "x"
    await put_task
    assert await q.get() == "y"
    assert await q.get() == "z"


def test_simple_async_queue():
    asyncio.run(async_test_simple_async_queue())



import asyncio
import time

import pytest

from storey import AsyncEmitSource
from storey.flow import Complete, ConcurrentExecution, Reduce, build_flow
from tests.test_flow import append_and_return

event_processing_duration = 0.5


class SomeContext:
    def __init__(self):
        self.fn = lambda x: x


async def process_event_slow_asyncio(event, context):
    assert isinstance(context, SomeContext) and callable(context.fn)
    await asyncio.sleep(event_processing_duration)
    return event


def process_event_slow_io(event, context):
    assert isinstance(context, SomeContext) and callable(context.fn)
    time.sleep(event_processing_duration)
    return event


def process_event_slow_processing(event):
    start = time.monotonic()
    while time.monotonic() - start < event_processing_duration:
        pass
    return event


async def async_test_concurrent_execution(concurrency_mechanism, event_processor, pass_context):
    controller = build_flow(
        [
            AsyncEmitSource(),
            ConcurrentExecution(
                event_processor=event_processor,
                concurrency_mechanism=concurrency_mechanism,
                pass_context=pass_context,
                max_in_flight=10,
                context=SomeContext(),
            ),
            Reduce([], append_and_return),
        ]
    ).run()

    num_events = 8

    start = time.monotonic()
    for counter in range(num_events):
        await controller.emit(counter)

    await controller.terminate()
    result = await controller.await_termination()
    end = time.monotonic()

    assert result == list(range(num_events))
    assert end - start > event_processing_duration, "Run time cannot be less than the time to process a single event"
    assert (
        end - start < event_processing_duration * num_events
    ), "Run time must be less than the time to process all events in serial"


@pytest.mark.parametrize(
    ["concurrency_mechanism", "event_processor", "pass_context"],
    [
        ("asyncio", process_event_slow_asyncio, True),
        ("threading", process_event_slow_io, True),
        ("multiprocessing", process_event_slow_processing, False),
    ],
)
def test_concurrent_execution(concurrency_mechanism, event_processor, pass_context):
    asyncio.run(async_test_concurrent_execution(concurrency_mechanism, event_processor, pass_context))


async def async_test_concurrent_execution_multiprocessing_and_complete():
    controller = build_flow(
        [
            AsyncEmitSource(),
            ConcurrentExecution(
                event_processor=process_event_slow_processing,
                concurrency_mechanism="multiprocessing",
                max_in_flight=2,
            ),
            Complete(),
        ]
    ).run()

    event_body = "hello"
    try:
        res = await controller.emit(event_body)
        assert res == event_body
    finally:
        await controller.terminate()
        await controller.await_termination()


def test_concurrent_execution_multiprocessing_and_complete():
    asyncio.run(async_test_concurrent_execution_multiprocessing_and_complete())


def test_concurrent_execution_multiprocessing_and_full_event():
    with pytest.raises(
        ValueError,
        match='concurrency_mechanism="multiprocessing" may not be used in conjunction with full_event=True',
    ):
        ConcurrentExecution(
            event_processor=process_event_slow_processing,
            concurrency_mechanism="multiprocessing",
            full_event=True,
        )



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__version__ = "1.7.28"

# Importing supported filesystems explicitly so that they will get registered as an fsspec filesystem
import v3iofs  # noqa: F401

from .aggregations import AggregateByKey, QueryByKey  # noqa: F401
from .dataframe import ReduceToDataFrame, ToDataFrame  # noqa: F401
from .drivers import Driver, NoopDriver, V3ioDriver  # noqa: F401
from .dtypes import EmissionType  # noqa: F401
from .dtypes import EmitAfterDelay  # noqa: F401
from .dtypes import EmitAfterMaxEvent  # noqa: F401
from .dtypes import EmitAfterPeriod  # noqa: F401
from .dtypes import EmitAfterWindow  # noqa: F401
from .dtypes import EmitEveryEvent  # noqa: F401
from .dtypes import EmitPolicy  # noqa: F401
from .dtypes import Event  # noqa: F401
from .dtypes import FieldAggregator  # noqa: F401
from .dtypes import FixedWindows  # noqa: F401
from .dtypes import FixedWindowType  # noqa: F401
from .dtypes import LateDataHandling  # noqa: F401
from .dtypes import SlidingWindows  # noqa: F401
from .flow import Batch  # noqa: F401
from .flow import Choice  # noqa: F401
from .flow import Complete  # noqa: F401
from .flow import ConcurrentExecution  # noqa: F401
from .flow import Context  # noqa: F401
from .flow import Extend  # noqa: F401
from .flow import Filter  # noqa: F401
from .flow import FlatMap  # noqa: F401
from .flow import Flow  # noqa: F401
from .flow import FlowError  # noqa: F401
from .flow import HttpRequest  # noqa: F401
from .flow import HttpResponse  # noqa: F401
from .flow import JoinWithTable  # noqa: F401
from .flow import JoinWithV3IOTable  # noqa: F401
from .flow import Map  # noqa: F401
from .flow import MapClass  # noqa: F401
from .flow import MapWithState  # noqa: F401
from .flow import Recover  # noqa: F401
from .flow import Reduce  # noqa: F401
from .flow import Rename  # noqa: F401
from .flow import SendToHttp  # noqa: F401
from .flow import build_flow  # noqa: F401
from .sources import AsyncEmitSource  # noqa: F401
from .sources import CSVSource  # noqa: F401
from .sources import DataframeSource  # noqa: F401
from .sources import ParquetSource  # noqa: F401
from .sources import SQLSource  # noqa: F401
from .sources import SyncEmitSource  # noqa: F401
from .sql_driver import SQLDriver  # noqa: F401
from .table import Table  # noqa: F401
from .targets import CSVTarget  # noqa: F401
from .targets import KafkaTarget  # noqa: F401
from .targets import NoSqlTarget  # noqa: F401
from .targets import ParquetTarget  # noqa: F401
from .targets import StreamTarget  # noqa: F401
from .targets import TDEngineTarget  # noqa: F401
from .targets import TSDBTarget  # noqa: F401

# clear module namespace
del v3iofs



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyncio
import copy
import gc
import queue
import threading
import time
import traceback
import warnings
import weakref
from collections import defaultdict
from datetime import datetime, timezone
from typing import Callable, Coroutine, Iterable, List, Optional, Union

import packaging.version
import pandas
import pandas as pd
import pyarrow
import pytz
from nuclio_sdk import QualifiedOffset

from .dtypes import Event, _termination_obj
from .flow import Complete, Flow, WithUUID
from .queue import SimpleAsyncQueue
from .utils import (
    find_filters,
    find_partitions,
    unpack_event_if_wrapped,
    url_to_file_system,
)


class AwaitableResult:
    """
    Future result of a computation. Calling await_result() will return with the result once the computation is
    completed.
    """

    def __init__(
        self,
        on_error: Optional[Callable[[], None]] = None,
        expected_number_of_results: int = 1,
    ):
        self._on_error = on_error
        self._expected_number_of_results = expected_number_of_results
        self._number_of_results = 0
        self._q = queue.Queue(expected_number_of_results)

    def await_result(self):
        """Returns the result, once the computation is completed"""
        results = []
        for _ in range(self._expected_number_of_results):
            result = self._q.get()
            if isinstance(result, BaseException):
                if self._on_error:
                    self._on_error()
                # Python appends trace frames to a raised exception, so we must copy
                # it before raising to prevent it from growing each time
                raise copy.copy(result)
            results.append(result)
        if len(results) == 1:
            results = results[0]
        return results

    def _set_result(self, element):
        if self._number_of_results < self._expected_number_of_results:
            self._number_of_results += 1
            self._q.put(element)

    def _set_error(self, ex):
        self._set_result(ex)


class FlowControllerBase(WithUUID):
    def __init__(
        self,
        key_field: Optional[Union[str, List[str]]],
        id_field: Optional[str],
    ):
        super().__init__()
        self._key_field = key_field
        self._id_field = id_field

    def _build_event(self, element, key):
        element_is_event = hasattr(element, "id")
        if element_is_event:
            element = unpack_event_if_wrapped(element)
            body = element.body
            if not hasattr(element, "processing_time"):
                if hasattr(element, "timestamp"):
                    element.processing_time = element.timestamp
                else:
                    element.processing_time = datetime.now(timezone.utc)
        else:
            body = element

        if isinstance(body, (dict, list)) and not key and self._key_field:
            if isinstance(self._key_field, str) or isinstance(self._key_field, int):
                key = body[self._key_field]
            else:
                key = []
                for field in self._key_field:
                    key.append(body[field])

        if element_is_event:
            if key or not hasattr(element, "key"):
                element.key = key
            return element
        else:
            return Event(body, id=self._get_uuid(), key=key)


class FlowController(FlowControllerBase):
    """Used to emit events into the associated flow, terminate the flow, and await the flow's termination.
    To be used from a synchronous context.
    """

    def __init__(
        self,
        emit_fn,
        await_termination_fn,
        return_awaitable_result,
        key_field: Optional[str] = None,
        id_field: Optional[str] = None,
    ):
        super().__init__(key_field, id_field)
        self._emit_fn = emit_fn
        self._await_termination_fn = await_termination_fn
        self._return_awaitable_result = return_awaitable_result

    def emit(
        self,
        element: object,
        key: Optional[Union[str, List[str]]] = None,
        return_awaitable_result: Optional[bool] = None,
        expected_number_of_results: Optional[int] = None,
    ):
        """Emits an event into the associated flow.

        :param element: The event data, or payload. To set metadata as well, pass an Event object.
        :param key: The event key(s) (optional) #add to async
        :param return_awaitable_result: Deprecated. An awaitable result object will be returned if a Complete step
            appears in the flow.
        :param expected_number_of_results: Number of times the event will have to pass through a Complete step to be
            completed (for graph flows).

        :returns: AsyncAwaitableResult if a Complete appears in the flow. None otherwise.
        """
        if return_awaitable_result is not None:
            warnings.warn(
                "return_awaitable_result is deprecated. An awaitable result object will be returned if a Complete step "
                "appears in the flow.",
                DeprecationWarning,
            )

        event = self._build_event(element, key)
        awaitable_result = None
        if self._return_awaitable_result:
            awaitable_result = AwaitableResult(expected_number_of_results=expected_number_of_results or 1)
        event._awaitable_result = awaitable_result
        event._original_events = [event]
        self._emit_fn(event)
        return awaitable_result

    def terminate(self):
        """Terminates the associated flow."""
        self._emit_fn(_termination_obj)

    def await_termination(self):
        """Awaits the termination of the flow. To be called after terminate. Returns the termination result of the
        flow (if any).
        """
        return self._await_termination_fn()


class FlowAwaiter:
    """Future termination result of a flow. Calling await_termination() will wait for the flow to terminate and return
    its termination result."""

    def __init__(self, await_termination_fn):
        self._await_termination_fn = await_termination_fn

    def await_termination(self):
        """ "waits for the flow to terminate and returns the result"""
        return self._await_termination_fn()


class _EventOffset:
    def __init__(self, event):
        self.event_weakref = weakref.ref(event)
        self.offset = event.offset

    def is_ready_to_commit(self):
        return self.event_weakref() is None

    def __repr__(self):
        return f"_EventOffset({self.offset})"


class SyncEmitSource(Flow):
    """Synchronous entry point into a flow. Produces a FlowController when run, for use from inside a synchronous
    context. See AsyncEmitSource for use from inside an async context.

    :param buffer_size: size of the incoming event buffer. Defaults to 8.
    :param key_field: Field to extract and use as the key. Optional.
    :param max_events_before_commit: Maximum number of events to be processed before committing offsets.
      Defaults to 20,000.
    :param max_time_before_commit: Maximum number of seconds before committing offsets. Defaults to 45.
    :param max_wait_before_commit: Maximum number of seconds to wait for an event before committing offsets.
    :param explicit_ack: Whether to explicitly commit offsets. Defaults to False.
    :param name: Name of this step, as it should appear in logs. Defaults to class name (SyncEmitSource).
    :type name: string

    for additional params, see documentation of  :class:`storey.flow.Flow`
    """

    _legal_first_step = True

    def __init__(
        self,
        buffer_size: Optional[int] = None,
        key_field: Union[list, str, int, None] = None,
        max_events_before_commit=None,
        max_time_before_commit=None,
        max_wait_before_commit=None,
        explicit_ack=False,
        **kwargs,
    ):
        if buffer_size is None:
            buffer_size = 8
        else:
            kwargs["buffer_size"] = buffer_size
        if key_field is not None:
            kwargs["key_field"] = key_field
        super().__init__(**kwargs)
        if buffer_size <= 0:
            raise ValueError("Buffer size must be positive")
        self._buffer_size = buffer_size
        self._key_field = key_field
        self._max_events_before_commit = max_events_before_commit or 20000
        self._max_time_before_commit = max_time_before_commit or 45
        self._max_wait_before_commit = max_wait_before_commit or 5
        self._explicit_ack = explicit_ack
        self._ex = None
        self._closeables = []

    def _init(self):
        super()._init()
        self._q = queue.Queue(self._buffer_size)
        self._termination_q = queue.Queue(1)
        self._is_terminated = False
        self._outstanding_offsets = defaultdict(list)

    async def _run_loop(self):
        loop = asyncio.get_running_loop()
        self._termination_future = loop.create_future()
        committer = None
        num_offsets_not_committed = 0
        events_handled_since_commit = 0
        last_commit_time = time.monotonic()
        if self._explicit_ack and hasattr(self.context, "platform") and hasattr(self.context.platform, "explicit_ack"):
            committer = self.context.platform.explicit_ack
        while True:
            event = None
            if committer:
                if (
                    events_handled_since_commit >= self._max_events_before_commit
                    or num_offsets_not_committed > 1
                    and time.monotonic() >= last_commit_time + self._max_time_before_commit
                ):
                    num_offsets_not_committed = await _commit_handled_events(self._outstanding_offsets, committer)
                    events_handled_since_commit = 0
                    last_commit_time = time.monotonic()
                # Due to the last event not being garbage collected, we tolerate a single unhandled event
                # TODO: Fix after transitioning to AsyncEmitSource, which would solve the underlying problem
                while num_offsets_not_committed > 1:
                    try:
                        event = await loop.run_in_executor(None, self._q.get, True, self._max_wait_before_commit)
                        break
                    except queue.Empty:
                        pass
                    num_offsets_not_committed = await _commit_handled_events(self._outstanding_offsets, committer)
                    events_handled_since_commit = 0
                    last_commit_time = time.monotonic()
            if event is None:
                event = await loop.run_in_executor(None, self._q.get)
            if committer and hasattr(event, "path") and hasattr(event, "shard_id") and hasattr(event, "offset"):
                # ML-6065 – workaround for NUC-178
                stream_path = event.stream_path if hasattr(event, "stream_path") else event.path
                qualified_shard = (stream_path, event.shard_id)
                offsets = self._outstanding_offsets[qualified_shard]
                offsets.append(_EventOffset(event))
                num_offsets_not_committed += 1
                events_handled_since_commit += 1
            try:
                termination_result = await self._do_downstream(event)
                if event is _termination_obj:
                    # We can commit all at this point because termination of
                    # all downstream steps completed successfully.
                    await _commit_handled_events(self._outstanding_offsets, committer, commit_all=True)
                    self._termination_future.set_result(termination_result)
            except BaseException as ex:
                if self.logger:
                    message = "An error was raised"
                    raised_by = getattr(ex, "_raised_by_storey_step", None)
                    if raised_by:
                        message += f" by step {type(raised_by)}"
                    self.logger.error(f"{message}: {traceback.format_exc()}")
                if event is not _termination_obj and event._awaitable_result:
                    event._awaitable_result._set_error(ex)
                self._ex = ex
                if not self._q.empty():
                    event = self._q.get()
                    if event is not _termination_obj and event._awaitable_result:
                        event._awaitable_result._set_error(ex)
                self._termination_future.set_result(None)
                break
            if event is _termination_obj:
                break

        for closeable in self._closeables:
            try:
                maybe_coroutine = closeable.close()
                if asyncio.iscoroutine(maybe_coroutine):
                    await maybe_coroutine
            except Exception as ex:
                if self.context:
                    self.logger.error(f"Error trying to close {closeable}: {ex}")

    def _loop_thread_main(self):
        asyncio.run(self._run_loop())
        self._termination_q.put(self._ex)

    def _raise_on_error(self, ex):
        if ex:
            # Python appends trace frames to a raised exception, so we must copy
            # it before raising to prevent it from growing each time
            ex_copy = copy.copy(self._ex)
            if self.verbose:
                raise type(ex_copy)("Flow execution terminated") from ex_copy
            raise ex_copy

    def _emit(self, event):
        if event is not _termination_obj:
            self._raise_on_error(self._ex)
            if self._is_terminated:
                raise ValueError("Cannot emit to a terminated flow")
        else:
            self._is_terminated = True
        self._q.put(event)
        if event is not _termination_obj:
            self._raise_on_error(self._ex)

    def run(self):
        """Starts the flow"""
        self._closeables = super().run()

        thread = threading.Thread(target=self._loop_thread_main)
        thread.start()

        def raise_error_or_return_termination_result():
            self._raise_on_error(self._termination_q.get())
            return self._termination_future.result()

        has_complete = self._check_step_in_flow(Complete)

        return FlowController(
            self._emit,
            raise_error_or_return_termination_result,
            has_complete,
            self._key_field,
        )


class AsyncAwaitableResult:
    """Future result of a computation. Calling await_result() will return with the result once the computation is
    completed. Same as AwaitableResult but for an async context.
    """

    def __init__(
        self,
        on_error: Optional[Callable[[BaseException], Coroutine]] = None,
        expected_number_of_results: int = 1,
    ):
        self._on_error = on_error
        self._expected_number_of_results = expected_number_of_results
        self._number_of_results = 0
        self._q = asyncio.Queue(expected_number_of_results)

    async def await_result(self):
        """returns the result of the computation, once the computation is complete"""
        results = []
        for _ in range(self._expected_number_of_results):
            result = await self._q.get()
            if isinstance(result, BaseException):
                if self._on_error:
                    await self._on_error()
                # Python appends trace frames to a raised exception, so we must copy
                # it before raising to prevent it from growing each time
                raise copy.copy(result)
            results.append(result)
        if len(results) == 1:
            results = results[0]
        return results

    async def _set_result(self, element):
        if self._number_of_results < self._expected_number_of_results:
            self._number_of_results += 1
            await self._q.put(element)

    async def _set_error(self, ex):
        await self._set_result(ex)


class AsyncFlowController(FlowControllerBase):
    """
    Used to emit events into the associated flow, terminate the flow, and await the flow's termination. To be used from
    inside an async def.
    """

    def __init__(
        self,
        emit_fn,
        loop_task,
        await_result,
        key_field: Optional[str] = None,
        id_field: Optional[str] = None,
    ):
        super().__init__(key_field, id_field)
        self._emit_fn = emit_fn
        self._loop_task = loop_task
        self._key_field = key_field
        self._await_result = await_result

    async def emit(
        self,
        element: object,
        key: Optional[Union[str, List[str]]] = None,
        await_result: Optional[bool] = None,
        expected_number_of_results: Optional[int] = None,
    ) -> object:
        """Emits an event into the associated flow.

        :param element: The event data, or payload. To set metadata as well, pass an Event object.
        :param key: The event key(s) (optional)
        :param await_result: Deprecated. Will await a result if a Complete step appears in the flow.
        :param expected_number_of_results: Number of times the event will have to pass through a Complete step to be
            completed (for graph flows).

        :returns: The result received from the flow if a Complete step appears in the flow. None otherwise.
        """
        if await_result is not None:
            warnings.warn(
                "await_result is deprecated. An awaitable result object will be returned if a Complete step appears "
                "in the flow.",
                DeprecationWarning,
            )

        event = self._build_event(element, key)
        awaitable = None
        if self._await_result:
            awaitable = AsyncAwaitableResult()
        event._awaitable_result = awaitable
        event._original_events = [event]
        await self._emit_fn(event)
        if self._await_result:
            result = await awaitable.await_result()
            if isinstance(result, BaseException):
                raise result
            return result

    async def terminate(self):
        """Terminates the associated flow."""
        await self._emit_fn(_termination_obj)

    async def await_termination(self):
        """
        Awaits the termination of the flow. To be called after terminate. Returns the termination result of the
        flow (if any).
        """
        return await self._loop_task


async def _commit_handled_events(outstanding_offsets_by_qualified_shard, committer, commit_all=False):
    num_offsets_not_handled = 0
    if not commit_all:
        gc.collect()
    for qualified_shard, offsets in outstanding_offsets_by_qualified_shard.items():
        if commit_all and offsets:
            last_handled_offset = offsets[-1].offset
            num_to_clear = len(offsets)
        else:
            num_to_clear = 0
            last_handled_offset = None
            # go over offsets in the qualified shard by arrival order until we reach an unhandled offset
            for i, offset in enumerate(offsets):
                if not offset.is_ready_to_commit():
                    num_offsets_not_handled += len(offsets) - i
                    break
                last_handled_offset = offset.offset
                num_to_clear += 1
        if last_handled_offset is not None:
            path, shard_id = qualified_shard
            await committer(QualifiedOffset(path, shard_id, last_handled_offset))
            outstanding_offsets_by_qualified_shard[qualified_shard] = offsets[num_to_clear:]
    return num_offsets_not_handled


class AsyncEmitSource(Flow):
    """
    Asynchronous entry point into a flow. Produces an AsyncFlowController when run, for use from inside an async def.
    See SyncEmitSource for use from inside a synchronous context.

    :param buffer_size: size of the incoming event buffer. Defaults to 8.
    :param key_field: Field to extract and use as the key. Optional.
    :param max_events_before_commit: Maximum number of events to be processed before committing offsets.
      Defaults to 20,000.
    :param max_time_before_commit: Maximum number of seconds before committing offsets. Defaults to 45.
    :param max_wait_before_commit: Maximum number of seconds to wait for an event before committing offsets.
    :param explicit_ack: Whether to explicitly commit offsets. Defaults to False.
    :param name: Name of this step, as it should appear in logs. Defaults to class name (AsyncEmitSource).
    :type name: string

    for additional params, see documentation of  :class:`~storey.flow.Flow`
    """

    _legal_first_step = True

    def __init__(
        self,
        buffer_size: int = None,
        key_field: Union[list, str, None] = None,
        max_events_before_commit=None,
        max_time_before_commit=None,
        max_wait_before_commit=None,
        explicit_ack=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if buffer_size is None:
            self._buffer_size = 8
        elif buffer_size <= 0:
            raise ValueError("Buffer size must be positive")
        else:
            kwargs["buffer_size"] = self._buffer_size
        self._key_field = key_field
        self._max_events_before_commit = max_events_before_commit or 20000
        self._max_time_before_commit = max_time_before_commit or 45
        self._max_wait_before_commit = max_wait_before_commit or 5
        self._explicit_ack = explicit_ack
        self._ex = None
        self._closeables = []

    def _init(self):
        super()._init()
        self._is_terminated = False
        self._outstanding_offsets = defaultdict(list)
        self._q = SimpleAsyncQueue(self._buffer_size)

    async def _run_loop(self):
        committer = None
        num_offsets_not_handled = 0
        events_handled_since_commit = 0
        last_commit_time = time.monotonic()
        if self._explicit_ack and hasattr(self.context, "platform") and hasattr(self.context.platform, "explicit_ack"):
            committer = self.context.platform.explicit_ack
        while True:
            event = None
            if committer:
                if (
                    events_handled_since_commit >= self._max_events_before_commit
                    or num_offsets_not_handled > 0
                    and time.monotonic() >= last_commit_time + self._max_time_before_commit
                ):
                    num_offsets_not_handled = await _commit_handled_events(self._outstanding_offsets, committer)
                    events_handled_since_commit = 0
                    last_commit_time = time.monotonic()
                # In case we can't block because there are outstanding events
                while num_offsets_not_handled > 0:
                    try:
                        event = await self._q.get(self._max_wait_before_commit)
                        break
                    except TimeoutError:
                        pass
                    num_offsets_not_handled = await _commit_handled_events(self._outstanding_offsets, committer)
                    events_handled_since_commit = 0
                    last_commit_time = time.monotonic()
            if not event:
                event = await self._q.get()
            if committer and hasattr(event, "path") and hasattr(event, "shard_id") and hasattr(event, "offset"):
                # ML-6065 – workaround for NUC-178
                stream_path = event.stream_path if hasattr(event, "stream_path") else event.path
                qualified_shard = (stream_path, event.shard_id)
                offsets = self._outstanding_offsets[qualified_shard]
                offsets.append(_EventOffset(event))
                num_offsets_not_handled += 1
                events_handled_since_commit += 1
            try:
                termination_result = await self._do_downstream(event)
                if event is _termination_obj:
                    # We can commit all at this point because termination of
                    # all downstream steps completed successfully.
                    await _commit_handled_events(self._outstanding_offsets, committer, commit_all=True)
                    return termination_result
            except BaseException as ex:
                if self.logger:
                    message = "An error was raised"
                    raised_by = getattr(ex, "_raised_by_storey_step", None)
                    if raised_by:
                        message += f" by step {type(raised_by)}"
                    self.logger.error(f"{message}: {traceback.format_exc()}")
                self._ex = ex
                if event is not _termination_obj and event._awaitable_result:
                    awaitable = event._awaitable_result._set_error(ex)
                    if awaitable:
                        await awaitable
                if not self._q.empty():
                    await self._q.get()
                self._raise_on_error()
            finally:
                if event is _termination_obj or self._ex:
                    for closeable in self._closeables:
                        try:
                            maybe_coroutine = closeable.close()
                            if asyncio.iscoroutine(maybe_coroutine):
                                await maybe_coroutine
                        except Exception as ex:
                            if self.context:
                                self.logger.error(f"Error trying to close {closeable}: {ex}")

    def _raise_on_error(self):
        if self._ex:
            # Python appends trace frames to a raised exception, so we must copy
            # it before raising to prevent it from growing each time
            ex_copy = copy.copy(self._ex)
            if self.verbose:
                raise type(ex_copy)("Flow execution terminated") from ex_copy
            raise ex_copy

    async def _emit(self, event):
        if event is not _termination_obj:
            self._raise_on_error()
            if self._is_terminated:
                raise ValueError("Cannot emit to a terminated flow")
        else:
            self._is_terminated = True
        await self._q.put(event)
        if event is not _termination_obj:
            self._raise_on_error()

    def run(self):
        """Starts the flow"""
        self._closeables = super().run()
        loop_task = asyncio.get_running_loop().create_task(self._run_loop())
        has_complete = self._check_step_in_flow(Complete)
        return AsyncFlowController(self._emit, loop_task, has_complete, self._key_field)


class _IterableSource(Flow):
    _legal_first_step = True

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._termination_q = queue.Queue(1)
        self._ex = None
        self._closeables = []

    async def _run_loop(self):
        raise NotImplementedError()

    async def _async_loop_thread_main(self):
        try:
            self._termination_future = asyncio.get_running_loop().create_future()
            termination_result = await self._run_loop()
            self._termination_future.set_result(termination_result)
        except BaseException as ex:
            self._ex = ex
            self._termination_future.set_result(None)
        finally:
            for closeable in self._closeables:
                await closeable.close()

    def _loop_thread_main(self):
        asyncio.run(self._async_loop_thread_main())
        self._termination_q.put(self._ex)

    def _raise_on_error(self, ex):
        if ex:
            if self.verbose:
                raise type(self._ex)("Flow execution terminated") from self._ex
            raise self._ex

    def run(self):
        self._closeables = super().run()

        self._init()

        thread = threading.Thread(target=self._loop_thread_main)
        thread.start()

        def raise_error_or_return_termination_result():
            self._raise_on_error(self._termination_q.get())
            return self._termination_future.result()

        return FlowAwaiter(raise_error_or_return_termination_result)

    async def run_async(self):
        self._closeables = super().run()
        return await self._run_loop()


class DataframeSource(_IterableSource, WithUUID):
    """Use pandas dataframe as input source for a flow.

    :param dfs: A pandas dataframe, or dataframes, to be used as input source for the flow.
    :param key_field: column to be used as key for events. can be list of columns
    :param id_field: column to be used as ID for events.

    for additional params, see documentation of  :class:`~storey.flow.Flow`
    """

    def __init__(
        self,
        dfs: Union[pandas.DataFrame, Iterable[pandas.DataFrame]],
        key_field: Optional[Union[str, List[str]]] = None,
        id_field: Optional[str] = None,
        **kwargs,
    ):
        if key_field is not None:
            kwargs["key_field"] = key_field
        if id_field is not None:
            kwargs["id_field"] = id_field
        _IterableSource.__init__(self, **kwargs)
        WithUUID.__init__(self)
        #  in order to raise exception also for key_field=0
        if key_field is not None:
            key_fields = [key_field] if not isinstance(key_field, list) else key_field
            if any([not isinstance(single_key_field, str) for single_key_field in key_fields]):
                raise ValueError("key_field must be a string or list of strings")
        if id_field and not isinstance(id_field, str):
            raise ValueError("id_field must be a string")
        self._key_field = key_field
        self._id_field = id_field
        if isinstance(dfs, pandas.DataFrame):
            dfs = [dfs]
        if dfs:
            for df in dfs:
                self._validate_fields(df)
        self._dfs = dfs

    def _get_keys(self, body: dict):
        keys = []
        if self._key_field:
            if isinstance(self._key_field, list):
                for key_field in self._key_field:
                    single_key = body[key_field]
                    keys.append(single_key)
            else:
                keys.append(body[self._key_field])
        return keys

    async def _run_loop(self):
        for df in self._dfs:
            columns = list(df.columns)
            is_df_index_nonempty = not df.index.empty and not (len(df.index.names) == 1 and df.index.names[0] is None)
            if is_df_index_nonempty:
                df = df.reset_index(drop=False)
            for body in df.to_dict("records"):
                keys = self._get_keys(body)
                key_fields = self._key_field if isinstance(self._key_field, list) else [self._key_field]
                none_keys = [key for key, value in zip(key_fields, keys) if pd.isna(value)]
                if none_keys:
                    if self.context:
                        self.logger.error(
                            f"Encountered null values in the following key fields:"
                            f" {', '.join(none_keys)}, in line: {body}."
                        )
                else:
                    if self._id_field:
                        line_id = body[self._id_field]
                    else:
                        line_id = self._get_uuid()
                    element = self._get_element(body, columns)
                    if len(keys) == 0:
                        keys = None
                    elif not isinstance(self._key_field, list):
                        keys = keys[0]
                    event = Event(element, keys, id=line_id)
                    await self._do_downstream(event)
        return await self._do_downstream(_termination_obj)

    def _validate_fields(self, df, path=""):
        path_message = f" File path: {path}." if path else ""
        df = df.reset_index()
        if self._key_field:
            key_fields = [self._key_field] if not isinstance(self._key_field, list) else self._key_field
            missing_keys = [key_field for key_field in key_fields if key_field not in df.columns]
            if missing_keys:
                if len(missing_keys) > 1:
                    missing_keys_message = f"key columns {missing_keys} are"
                else:
                    missing_keys_message = f"key column '{missing_keys[0]}' is"
                raise ValueError(f"{missing_keys_message} missing from dataframe.{path_message}")

        if self._id_field and self._id_field not in df.columns:
            raise ValueError(f"id column '{self._id_field}' is missing from dataframe.{path_message}")

    def _get_element(self, body: dict, columns):
        return body


class CSVSource(DataframeSource):
    """
    Reads CSV files as input source for a flow.

    :parameter paths: paths to CSV files
    :parameter header: Deprecated. whether CSV files have a header or not. Defaults to False.
    :parameter build_dict: whether to format each record produced from the input file as a dictionary (as opposed to a
        list). Default to False.
    :parameter key_field: the CSV field to be used as the key for events. May be an int (field index) or string (field
        name) if with_header is True. Defaults to None (no key). Can be a list of keys
    :parameter time_field: the CSV field to be parsed as the timestamp for events. May be an int (field index) or string
        (field name) if with_header is True. Defaults to None (no timestamp field).
    :parameter timestamp_format: timestamp format as defined in datetime.strptime(). Default to ISO-8601 as defined in
        datetime.fromisoformat().
    :parameter id_field: the CSV field to be used as the ID for events. May be an int (field index) or string (field
        name) if with_header is True. Defaults to None (random ID will be generated per event).
    :parameter type_inference: Deprecated. Whether to infer data types from the data (when True), or read all fields in
     as strings (when False). Defaults to True.
    :parameter parse_dates: list of columns (names or integers) that will be attempted to parse as date column

    for additional params, see documentation of  :class:`~storey.flow.Flow`
    """

    def __init__(
        self,
        paths: Union[List[str], str],
        header: bool = True,
        build_dict: bool = False,
        key_field: Union[int, str, List[int], List[str], None] = None,
        time_field: Union[int, str, None] = None,
        timestamp_format: Optional[str] = None,
        id_field: Union[str, int, None] = None,
        type_inference: bool = True,
        parse_dates: Optional[Union[int, str, List[int], List[str]]] = None,
        **kwargs,
    ):

        kwargs["paths"] = paths
        if isinstance(paths, str):
            paths = [paths]
        self._paths = paths
        self._build_dict = build_dict
        if header is False:
            warnings.warn(
                "header=False is deprecated, will be treated as header=True."
                " The parameter will be removed in a future version.",
                DeprecationWarning,
            )
        if type_inference is False:
            warnings.warn(
                "type_inference is deprecated, will be treated as type_inference=True."
                " The parameter will be removed in a future version.",
                DeprecationWarning,
            )
        kwargs["header"] = header
        kwargs["build_dict"] = build_dict
        if key_field is not None:
            kwargs["key_field"] = key_field
        if time_field is not None:
            kwargs["time_field"] = time_field
        if id_field is not None:
            kwargs["id_field"] = id_field
        if timestamp_format is not None:
            kwargs["timestamp_format"] = timestamp_format
        kwargs["type_inference"] = type_inference
        self._storage_options = kwargs.get("storage_options")
        self._paths = paths
        self._key_field = key_field
        self._time_field = time_field
        self._timestamp_format = timestamp_format
        self._id_field = id_field
        self._type_inference = type_inference
        self._storage_options = kwargs.get("storage_options")
        self._parse_dates = parse_dates
        self._dates_indices = []
        if parse_dates:
            if not isinstance(parse_dates, List):
                parse_dates = [parse_dates]
            self._dates_indices.extend(parse_dates)
        if time_field is not None:
            self._dates_indices.append(time_field)

        super().__init__([], **kwargs)

    def _init(self):
        super()._init()
        self._dfs = []
        for path in self._paths:
            kwargs = {}
            if packaging.version.Version(pandas.__version__).major >= 2:
                kwargs["date_format"] = self._timestamp_format
            else:
                kwargs["date_parser"] = self._datetime_from_timestamp
            df = pandas.read_csv(
                path,
                header=0,
                parse_dates=self._dates_indices,
                storage_options=self._storage_options,
                **kwargs,
            )
            self._validate_fields(df, path)
            self._dfs.append(df)

    def _datetime_from_timestamp(self, timestamp):
        if timestamp == "" or pd.isna(timestamp):
            return None
        if self._timestamp_format:
            return pandas.to_datetime(timestamp, format=self._timestamp_format).floor("u").to_pydatetime()
        else:
            return datetime.fromisoformat(timestamp)

    def _get_element(self, body: dict, columns: List[str]):
        if self._build_dict:
            return body
        return [body[column] for column in columns]


class ParquetSource(DataframeSource):
    """Reads Parquet files as input source for a flow.

    :parameter paths: paths to Parquet files
    :parameter columns: list, default=None. If not None, only these columns will be read from the file.
    :parameter start_filter: datetime. If not None, the results will be filtered by partitions and
        'filter_column' > start_filter. Default is None.
    :parameter end_filter: datetime. If not None, the results will be filtered by partitions
        'filter_column' <= end_filter. Default is None.
    :parameter filter_column: Optional. if not None, the results will be filtered by this column and before and/or after
                datetime column.
    :param key_field: column to be used as key for events. can be list of columns
    :param id_field: column to be used as ID for events.
    :param additional_filters: other filters to use while reading the parquet.
                    Supported operators: '=', '>=', '<=', '>', '<'.
                    Example: ('Product', '=', 'Computer')]
    For all supported filters, please see:
    https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetDataset.html
    """

    def __init__(
        self,
        paths: Union[str, Iterable[str]],
        columns=None,
        start_filter: Optional[datetime] = None,
        end_filter: Optional[datetime] = None,
        filter_column: Optional[str] = None,
        additional_filters: Optional[list[tuple]] = None,
        **kwargs,
    ):
        if start_filter or end_filter:
            if start_filter is None:
                start_filter = datetime.min
                if end_filter.tzinfo:
                    start_filter = start_filter.replace(tzinfo=pytz.utc)
            if end_filter is None:
                end_filter = datetime.max
                if start_filter.tzinfo:
                    end_filter = end_filter.replace(tzinfo=pytz.utc)
            if (start_filter.tzinfo is None) ^ (end_filter.tzinfo is None):
                raise ValueError("Start and end filters must either both have a time zone or both be naive timestamps")

            if filter_column is None:
                raise TypeError("Filter column is required when passing start/end filters")
        additional_filters = additional_filters or []

        if not all(isinstance(item, tuple) for item in additional_filters):
            raise ValueError(
                f"ParquetSource supports additional_filters only as a list of tuples."
                f" Current additional_filters: {additional_filters}"
            )

        self._paths = paths
        if isinstance(paths, str):
            self._paths = [paths]
        self._columns = columns
        self._start_filter = start_filter
        self._end_filter = end_filter
        self._filter_column = filter_column
        self._storage_options = kwargs.get("storage_options")
        self._additional_filters = additional_filters
        super().__init__([], **kwargs)

    def _read_filtered_parquet(self, path):
        fs, file_path = url_to_file_system(path, self._storage_options)

        partitions_time_attributes = find_partitions(path, fs)
        filters = []
        find_filters(
            partitions_time_attributes,
            self._start_filter,
            self._end_filter,
            filters,
            self._filter_column,
        )
        if filters and self._additional_filters:
            filters[0] += self._additional_filters
        try:
            return pandas.read_parquet(
                path,
                columns=self._columns,
                filters=filters,
                storage_options=self._storage_options,
            )
        except pyarrow.lib.ArrowInvalid as ex:
            if not str(ex).startswith("Cannot compare timestamp with timezone to timestamp without timezone"):
                raise ex

            if self._start_filter.tzinfo:
                start_filter = self._start_filter.replace(tzinfo=None)
                end_filter = self._end_filter.replace(tzinfo=None)
            else:
                start_filter = self._start_filter.replace(tzinfo=pytz.utc)
                end_filter = self._end_filter.replace(tzinfo=pytz.utc)

            filters = []
            find_filters(
                partitions_time_attributes,
                start_filter,
                end_filter,
                filters,
                self._filter_column,
            )
            if filters and self._additional_filters:
                filters[0] += self._additional_filters

            return pandas.read_parquet(
                path,
                columns=self._columns,
                filters=filters,
                storage_options=self._storage_options,
            )

    def _init(self):
        super()._init()
        self._dfs = []
        for path in self._paths:
            if self._start_filter or self._end_filter or self._additional_filters:
                df = self._read_filtered_parquet(path)
            else:
                df = pandas.read_parquet(path, columns=self._columns, storage_options=self._storage_options)
            self._validate_fields(df, path)
            self._dfs.append(df)


class SQLSource(_IterableSource, WithUUID):
    """Use SQL table as input source for a flow.

    :parameter key_field: the primary key of the table.
    :parameter id_field: column to be used as ID for events.
    :parameter db_path: url string connection to sql database.
    :parameter table_name: the name of the table to access, from the current database
    :parameter time_fields: list of all fields that are timestamps
    """

    def __init__(
        self,
        db_path: str,
        table_name: str,
        key_field: Union[None, str, List[str]] = None,
        id_field: str = None,
        time_fields: List[str] = None,
        **kwargs,
    ):

        if key_field is not None:
            kwargs["key_field"] = key_field
        if id_field is not None:
            kwargs["id_field"] = id_field
        _IterableSource.__init__(self, **kwargs)
        WithUUID.__init__(self)

        self.table_name = table_name
        self.db_path = db_path
        self.time_fields = time_fields

        self._key_field = key_field
        self._id_field = id_field

    async def _run_loop(self):
        import sqlalchemy as db

        engine = db.create_engine(self.db_path)
        with engine.connect() as conn:
            metadata = db.MetaData()

            table = db.Table(
                self.table_name,
                metadata,
                autoload=True,
                autoload_with=engine,
            )
            select_object = db.select(table)
            cursor = pandas.read_sql(select_object, con=conn, parse_dates=self.time_fields, chunksize=100)

            for df in cursor:
                for row in df.itertuples(index=False):
                    body = dict(row._asdict())
                    key = None
                    if self._key_field:
                        if isinstance(self._key_field, list):
                            key = []
                            for key_field in self._key_field:
                                if key_field not in body or pandas.isna(body[key_field]):
                                    self.logger.error(
                                        f"For {body} value there is no {self._key_field} " f"field (key_field)"
                                    )
                                    break
                                key.append(body[key_field])
                        else:
                            key = body.get(self._key_field, None)
                            if key is None:
                                self.logger.error(f"For {body} value there is no {self._key_field} field (key_field)")
                    if self._id_field:
                        event_id = body[self._id_field]
                    else:
                        event_id = self._get_uuid()
                    event = Event(body, key=key, id=event_id)
                    await self._do_downstream(event)
        return await self._do_downstream(_termination_obj)



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from unittest.mock import Mock

import pytest
import taosws

from storey.dtypes import TDEngineValueError
from storey.targets import TDEngineTarget


class TestTDEngineTarget:
    @staticmethod
    def test_tags_mapping_consistency() -> None:
        for type_, func in TDEngineTarget._get_tdengine_type_to_tag_func().items():
            assert func.__name__ == f"{type_.lower()}_to_tag"

    @staticmethod
    def test_columns_mapping_consistency() -> None:
        for type_, func in TDEngineTarget._get_tdengine_type_to_column_func().items():
            if type_ == "TIMESTAMP":
                assert func.__name__.startswith("millis_timestamp")
            else:
                assert func.__name__.startswith(type_.lower())
            assert func.__name__.endswith("_to_column")

    @staticmethod
    @pytest.mark.parametrize(
        ("database", "table", "supertable", "table_col", "tag_cols"),
        [
            (None, None, "my_super_tb", "pass_this_check", ["also_this_one"]),
            ("mydb", None, "my super  tb", "pass_this_check", ["also_this_one"]),
            ("_db", "9table", None, None, None),
            ("_db", " cars", None, None, None),
        ],
    )
    def test_invalid_names(
        database: Optional[str],
        table: Optional[str],
        supertable: Optional[str],
        table_col: Optional[str],
        tag_cols: Optional[list[str]],
    ) -> None:
        with pytest.raises(TDEngineValueError):
            TDEngineTarget(
                url="taosws://root:taosdata@localhost:6041",
                time_col="ts",
                columns=["value"],
                table_col=table_col,
                tag_cols=tag_cols,
                database=database,
                table=table,
                supertable=supertable,
            )

    @staticmethod
    @pytest.fixture
    def tdengine_target() -> TDEngineTarget:
        target = TDEngineTarget(
            url="taosws://root:taosdata@localhost:6041",
            time_col="ts",
            columns=["value"],
            database="test",
            table="d6241",
        )

        target._connection = Mock()
        # The following test schema is obtained from the `taosBenchmark` data:
        # https://docs.tdengine.com/get-started/docker/#test-data-insert-performance
        # list(conn.query("describe test.d6241;"))
        target._connection.query = Mock(
            return_value=[
                ("ts", "TIMESTAMP", 8, "", "delta-i", "lz4", "medium"),
                ("current", "FLOAT", 4, "", "delta-d", "lz4", "medium"),
                ("voltage", "INT", 4, "", "simple8b", "lz4", "medium"),
                ("phase", "FLOAT", 4, "", "delta-d", "lz4", "medium"),
                ("groupid", "INT", 4, "TAG", "disabled", "disabled", "disabled"),
                ("location", "VARCHAR", 24, "TAG", "disabled", "disabled", "disabled"),
            ],
        )
        return target

    @staticmethod
    def test_get_table_schema(tdengine_target: TDEngineTarget) -> None:
        """Test that the parsing works"""
        tags_schema, reg_cols_schema = tdengine_target._get_table_schema("d6241")
        assert tags_schema == [("groupid", taosws.int_to_tag), ("location", taosws.varchar_to_tag)]
        assert reg_cols_schema == [
            ("ts", taosws.millis_timestamps_to_column),
            ("current", taosws.floats_to_column),
            ("voltage", taosws.ints_to_column),
            ("phase", taosws.floats_to_column),
        ]



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyncio
import re
import time
from datetime import datetime
from typing import Callable, Dict, List, Optional, Union

import pandas as pd

from .aggregation_utils import is_aggregation_name
from .dtypes import (
    EmitAfterMaxEvent,
    EmitAfterPeriod,
    EmitAfterWindow,
    EmitEveryEvent,
    EmitPolicy,
    FieldAggregator,
    FixedWindows,
    FixedWindowType,
    SlidingWindows,
    _dict_to_emit_policy,
)
from .flow import Event, Flow, _termination_obj
from .table import Table
from .utils import stringify_key

_default_emit_policy = EmitEveryEvent()


class AggregateByKey(Flow):
    """
    Aggregates the data into the table object provided for later persistence,
    and outputs an event enriched with the requested aggregation features.
    Persistence is done via the `NoSqlTarget` step and based on the Cache object persistence settings.

    :param aggregates: List of aggregates to apply for each event.
        accepts either list of FieldAggregators or a dictionary describing FieldAggregators.
    :param table: A Table object or name for persistence of aggregations.
        If a table name is provided, it will be looked up in the context object passed in kwargs.
    :param key_field: Key field to aggregate by, accepts either a string representing the key field or a key
        extracting function. Defaults to the key in the event's metadata. (Optional)
    :param time_field: Time field to aggregate by, accepts either a string representing the time field or a time
        extracting function. Defaults to the processing time in the event's metadata. (Optional)
    :param emit_policy: Policy indicating when the data will be emitted. Defaults to EmitEveryEvent
    :param augmentation_fn: Function that augments the features into the event's body. Defaults to
        updating a dict. (Optional)
    :param enrich_with: List of attributes names from the associated storage object to be fetched
        and added to every event. (Optional)
    :param aliases: Dictionary specifying aliases for enriched or aggregate columns, of the
        format `{'col_name': 'new_col_name'}`. (Optional)
    :param time_format: If the value of the time field is of type string, this format will be used to parse it, as
        defined in datetime.strptime(). By default, parsing will follow ISO-8601.
    """

    def __init__(
        self,
        aggregates: Union[List[FieldAggregator], List[Dict[str, object]]],
        table: Union[Table, str],
        key_field: Union[str, List[str], Callable[[Event], object], None] = None,
        time_field: Union[str, Callable[[Event], object], None] = None,
        emit_policy: Union[EmitPolicy, Dict[str, object]] = _default_emit_policy,
        augmentation_fn: Optional[Callable[[Event, Dict[str, object]], Event]] = None,
        enrich_with: Optional[List[str]] = None,
        aliases: Optional[Dict[str, str]] = None,
        use_windows_from_schema: bool = False,
        time_format: Optional[str] = None,
        **kwargs,
    ):
        Flow.__init__(self, **kwargs)
        aggregates = self._parse_aggregates(aggregates)
        self._check_unique_names(aggregates)

        self._table = table
        if isinstance(table, str):
            if not self.context:
                raise TypeError("Table can not be string if no context was provided to the step")
            self._table = self.context.get_table(table)
        self._table._set_aggregation_metadata(aggregates, use_windows_from_schema=use_windows_from_schema)
        self._closeables = [self._table]

        self._aggregates_metadata = aggregates

        self._enrich_with = enrich_with or []
        self._aliases = aliases or {}

        if not isinstance(emit_policy, EmitPolicy) and not isinstance(emit_policy, Dict):
            raise TypeError(
                f"emit_policy parameter must be of type EmitPolicy, or dict. Found {type(emit_policy)} instead."
            )

        self._emit_policy = emit_policy
        if isinstance(self._emit_policy, dict):
            self._emit_policy = _dict_to_emit_policy(self._emit_policy)

        self._augmentation_fn = augmentation_fn
        if not augmentation_fn:

            def f(element, features):
                features.update(element)
                return features

            self._augmentation_fn = f

        self._key_extractor = None
        if key_field:
            if callable(key_field):
                self._key_extractor = key_field
            elif isinstance(key_field, str):
                self._key_extractor = lambda event: event.get(key_field)
            elif isinstance(key_field, list):
                self._key_extractor = lambda event: [event.get(single_key) for single_key in key_field]
            else:
                raise TypeError(
                    f"key_field is expected to be either a callable or string or list of strings "
                    f"but got {type(key_field)}"
                )

        self._time_extractor = lambda event: event.processing_time
        if time_field:
            if callable(time_field):
                self._time_extractor = time_field
            elif isinstance(time_field, str):
                self._time_extractor = lambda event: event.body.get(time_field)
            else:
                raise TypeError(f"time_field is expected to be either a callable or string but got {type(time_field)}")

        self._time_format = time_format

    def _init(self):
        super()._init()
        self._events_in_batch = {}
        self._emit_worker_running = False
        self._terminate_worker = False
        self._timeout_task: Optional[asyncio.Task] = None

    def _check_unique_names(self, aggregates):
        unique_aggr_names = set()
        for aggr in aggregates:
            if aggr.name in unique_aggr_names:
                raise TypeError(f"Aggregates should have unique names. {aggr.name} already exists")
            unique_aggr_names.add(aggr.name)

    @staticmethod
    def _parse_aggregates(aggregates):
        if not isinstance(aggregates, list):
            raise TypeError("aggregates should be a list of FieldAggregator/dictionaries")

        if not aggregates or isinstance(aggregates[0], FieldAggregator):
            return aggregates

        if isinstance(aggregates[0], dict):
            new_aggregates = []
            for aggregate_dict in aggregates:
                if "period" in aggregate_dict:
                    window = SlidingWindows(aggregate_dict["windows"], aggregate_dict["period"])
                else:
                    window = FixedWindows(aggregate_dict["windows"])
                new_aggregates.append(
                    FieldAggregator(
                        aggregate_dict["name"],
                        aggregate_dict["column"],
                        aggregate_dict["operations"],
                        window,
                        aggregate_dict.get("aggregation_filter", None),
                        aggregate_dict.get("max_value", None),
                    )
                )
            return new_aggregates

        raise TypeError("aggregates should be a list of FieldAggregator/dictionaries")

    def _get_timestamp(self, event):
        event_timestamp = self._time_extractor(event)
        if isinstance(event_timestamp, str):
            if self._time_format:
                event_timestamp = datetime.strptime(event_timestamp, self._time_format)
            else:
                event_timestamp = datetime.fromisoformat(event_timestamp)
        if isinstance(event_timestamp, datetime):
            if isinstance(event_timestamp, pd.Timestamp) and event_timestamp.tzinfo is None:
                # timestamp for pandas timestamp gives the wrong result in case there is no timezone (ML-313)
                local_time_zone = datetime.now().astimezone().tzinfo
                event_timestamp = event_timestamp.replace(tzinfo=local_time_zone)
            event_timestamp = event_timestamp.timestamp() * 1000
        return event_timestamp

    async def _do(self, event):
        if event == _termination_obj:
            self._terminate_worker = True
            return await self._do_downstream(_termination_obj)

        try:
            # check whether a background loop is needed, if so create start one
            if (not self._emit_worker_running) and (
                isinstance(self._emit_policy, EmitAfterPeriod) or isinstance(self._emit_policy, EmitAfterWindow)
            ):
                asyncio.get_running_loop().create_task(self._emit_worker())
                self._emit_worker_running = True

            element = event.body
            key = event.key
            if self._key_extractor:
                key = self._key_extractor(element)

            event_timestamp = self._get_timestamp(event)

            safe_key = stringify_key(key)
            await self._table._lazy_load_key_with_aggregates(safe_key, event_timestamp)
            await self._table._aggregate(safe_key, event, element, event_timestamp)

            if isinstance(self._emit_policy, EmitEveryEvent):
                await self._emit_event(key, event)
            elif isinstance(self._emit_policy, EmitAfterMaxEvent):
                if safe_key in self._events_in_batch:
                    self._events_in_batch[safe_key]["counter"] += 1
                else:
                    event_dict = {"counter": 1, "time": time.monotonic()}
                    self._events_in_batch[safe_key] = event_dict
                self._events_in_batch[safe_key]["event"] = event
                if self._emit_policy.timeout_secs and self._timeout_task is None:
                    self._timeout_task = asyncio.get_running_loop().create_task(self._sleep_and_emit())
                if self._events_in_batch[safe_key]["counter"] == self._emit_policy.max_events:
                    event_from_batch = self._events_in_batch.pop(safe_key, None)
                    if event_from_batch is not None:
                        await self._emit_event(key, event_from_batch["event"])
        except Exception as ex:
            raise ex

    async def _sleep_and_emit(self):
        while self._events_in_batch:
            key = next(iter(self._events_in_batch.keys()))
            delta_seconds = time.monotonic() - self._events_in_batch[key]["time"]
            if delta_seconds < self._emit_policy.timeout_secs:
                await asyncio.sleep(self._emit_policy.timeout_secs - delta_seconds)
            event = self._events_in_batch.pop(key, None)
            if event is not None:
                await self._emit_event(key, event["event"])

        self._timeout_task = None

    # Emit a single event for the requested key
    async def _emit_event(self, key, event):
        event_timestamp = self._get_timestamp(event)

        safe_key = stringify_key(key)
        await self._table._lazy_load_key_with_aggregates(safe_key, event_timestamp)
        features = await self._table._get_features(safe_key, event_timestamp)
        for feature_name in list(features.keys()):
            if feature_name in self._aliases:
                new_feature_name = self._aliases[feature_name]
                if feature_name != new_feature_name:
                    features[new_feature_name] = features[feature_name]
                    del features[feature_name]
        features = self._augmentation_fn(event.body, features)

        for col in self._enrich_with:
            emitted_attr_name = self._aliases.get(col, None) or col
            if col in self._table._get_static_attrs(safe_key):
                features[emitted_attr_name] = self._table._get_static_attrs(safe_key)[col]
        event.key = key
        event.body = features
        await self._do_downstream(event)

    # Emit multiple events for every key in the store with the current time
    async def _emit_all_events(self, timestamp):
        for key in self._table._get_keys():
            await self._emit_event(key, Event({"key": key, "time": timestamp}, key, timestamp, None))

    async def _emit_worker(self):
        if isinstance(self._emit_policy, EmitAfterPeriod):
            seconds_to_sleep_between_emits = self._aggregates_metadata[0].windows.period_millis / 1000
        elif isinstance(self._emit_policy, EmitAfterWindow):
            seconds_to_sleep_between_emits = self._aggregates_metadata[0].windows.windows[0][0] / 1000
        else:
            raise TypeError(f'Emit policy "{type(self._emit_policy)}" is not supported')

        current_time = datetime.now().timestamp()
        next_emit_time = (
            int(current_time / seconds_to_sleep_between_emits) * seconds_to_sleep_between_emits
            + seconds_to_sleep_between_emits
        )

        while not self._terminate_worker:
            current_time = datetime.now().timestamp()
            next_sleep_interval = next_emit_time - current_time + self._emit_policy.delay_in_seconds
            if next_sleep_interval > 0:
                await asyncio.sleep(next_sleep_interval)
            await self._emit_all_events(next_emit_time * 1000)
            next_emit_time = next_emit_time + seconds_to_sleep_between_emits


class QueryByKey(AggregateByKey):
    """
    Query features by name

    :param features: List of features to get.
    :param table: A Table object or name for persistence of aggregations.
        If a table name is provided, it will be looked up in the context object passed in kwargs.
    :param key_field: Key field to query by, accepts either a string representing the key field or a key
        extracting function. Defaults to the key in the event's metadata. Can be list of keys (Optional)
    :param time_field: Time field to query by, accepts either a string representing the time field or a time
        extracting function. Defaults to the processing time in the event's metadata. (Optional)
    :param augmentation_fn: Function that augments the features into the event's body.
        Defaults to updating a dict. (Optional)
    :param aliases: Dictionary specifying aliases for enriched or aggregate columns, of the
        format `{'col_name': 'new_col_name'}`. (Optional)
    :param options: Enum flags specifying query options. (Optional)
    """

    def __init__(
        self,
        features: List[str],
        table: Union[Table, str],
        key_field: Union[str, List[str], Callable[[Event], object], None] = None,
        time_field: Union[str, List[str], Callable[[Event], object], None] = None,
        augmentation_fn: Optional[Callable[[Event, Dict[str, object]], Event]] = None,
        aliases: Optional[Dict[str, str]] = None,
        fixed_window_type: Optional[FixedWindowType] = FixedWindowType.CurrentOpenWindow,
        **kwargs,
    ):
        self._aggrs = []
        self._enrich_cols = []
        resolved_aggrs = {}
        if isinstance(table, str):
            if "context" not in kwargs:
                raise TypeError("Table can not be string if no context was provided to the step")
            table = kwargs["context"].get_table(table)
        for feature in features:
            match = re.match(r".*_([a-z]+)_[0-9]+[smhd]$", feature) if table.supports_aggregations() else None
            if match and is_aggregation_name(match.group(1)):
                name, window = feature.rsplit("_", 1)
                if name in resolved_aggrs:
                    resolved_aggrs[name].append(window)
                else:
                    resolved_aggrs[name] = [window]
            else:
                self._enrich_cols.append(feature)
        for name, windows in resolved_aggrs.items():
            feature, aggr = name.rsplit("_", 1)
            # setting as SlidingWindow temporarily until actual window type will be read from schema
            self._aggrs.append(
                FieldAggregator(
                    name=feature,
                    field=None,
                    aggr=[aggr],
                    windows=SlidingWindows(windows),
                )
            )
        other_table = table._clone() if table._aggregates is not None else table
        AggregateByKey.__init__(
            self,
            self._aggrs,
            other_table,
            key_field,
            time_field,
            augmentation_fn=augmentation_fn,
            enrich_with=self._enrich_cols,
            aliases=aliases,
            use_windows_from_schema=True,
            **kwargs,
        )
        self._table._aggregations_read_only = True
        self._table.fixed_window_type = fixed_window_type

    async def _do(self, event):
        if event == _termination_obj:
            self._terminate_worker = True
            return await self._do_downstream(_termination_obj)

        element = event.body
        key = event.key
        if self._key_extractor:
            if element:
                key = self._key_extractor(element)
            if key is None or key == [None] or element is None:
                event.body = None
                await self._do_downstream(event)
                return
        await self._emit_event(key, event)

    def _check_unique_names(self, aggregates):
        pass



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from enum import Enum
from typing import Callable, Optional, Union

from storey import Flow
from storey.dtypes import Event, _termination_obj


class EmitPeriod(Enum):
    FIRST = 1
    LAST = 2


class SampleWindow(Flow):
    """
    Emits a single event in a window of `window_size` events, in accordance with `emit_period` and
    `emit_before_termination`.

    :param window_size: The size of the window we want to sample a single event from.
    :param emit_period: What event should this step emit for each `window_size` (default: EmitPeriod.First).
        Available options:
        1.1) EmitPeriod.FIRST - will emit the first event in a window `window_size` events.
        1.2) EmitPeriod.LAST - will emit the last event in a window of `window_size` events.
    :param emit_before_termination: On termination signal, should the step emit the last event it seen (default: False).
        Available options:
        2.1) True - The last event seen will be emitted downstream.
        2.2) False - The last event seen will NOT be emitted downstream.
    :param key: The key by which events are sampled. By default (None), events are not sampled by key.
        Other options may be:
        Set to '$key' to sample events by the Event.key property.
        set to 'str' key to sample events by Event.body[str].
        set a Callable[[Event], str] to sample events by a custom key extractor.
    """

    def __init__(
        self,
        window_size: int,
        emit_period: EmitPeriod = EmitPeriod.FIRST,
        emit_before_termination: bool = False,
        key: Optional[Union[str, Callable[[Event], str]]] = None,
        **kwargs,
    ):
        kwargs["full_event"] = True
        super().__init__(**kwargs)

        if window_size <= 1:
            raise ValueError(f"Expected window_size > 1, found {window_size}")

        if not isinstance(emit_period, EmitPeriod):
            raise ValueError(f"Expected emit_period of type `EmitPeriod`, got {type(emit_period)}")

        self._window_size = window_size
        self._emit_period = emit_period
        self._emit_before_termination = emit_before_termination
        self._per_key_count = dict()
        self._count = 0
        self._last_event = None
        self._extract_key: Callable[[Event], str] = self._create_key_extractor(key)

    @staticmethod
    def _create_key_extractor(key) -> Callable:
        if key is None:
            return lambda event: None
        elif callable(key):
            return key
        elif isinstance(key, str):
            if key == "$key":
                return lambda event: event.key
            else:
                return lambda event: event.body[key]
        else:
            raise ValueError(f"Unsupported key type {type(key)}")

    async def _do(self, event):
        if event is _termination_obj:
            if self._last_event is not None:
                await self._do_downstream(self._last_event)
            return await self._do_downstream(_termination_obj)
        else:
            key = self._extract_key(event)

            if key is not None:
                if key not in self._per_key_count:
                    self._per_key_count[key] = 1
                else:
                    self._per_key_count[key] += 1
                count = self._per_key_count[key]
            else:
                self._count += 1
                count = self._count

            if self._emit_before_termination:
                self._last_event = event

            if count == self._window_size:
                if key is not None:
                    self._per_key_count[key] = 0
                else:
                    self._count = 0
            if self._should_emit(count):
                self._last_event = None
                await self._do_downstream(event)

    def _should_emit(self, count):
        if self._emit_period == EmitPeriod.FIRST and count == 1:
            return True
        elif self._emit_period == EmitPeriod.LAST and count == self._window_size:
            return True
        return False



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyncio
import copy
import datetime
import inspect
import pickle
import time
import traceback
import uuid
from asyncio import Task
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Union

import aiohttp

from .dtypes import Event, FlowError, V3ioError, _termination_obj, known_driver_schemes
from .queue import AsyncQueue
from .table import Table
from .utils import _split_path, get_in, stringify_key, update_in


class Flow:
    _legal_first_step = False

    def __init__(
        self,
        recovery_step=None,
        termination_result_fn=lambda x, y: x if x is not None else y,
        context=None,
        **kwargs,
    ):
        self._outlets = []
        self._inlets = []

        self._recovery_step = recovery_step
        if recovery_step:
            if isinstance(recovery_step, dict):
                for step in recovery_step.values():
                    step._inlets.append(self)
            else:
                recovery_step._inlets.append(self)

        self._termination_result_fn = termination_result_fn
        self.context = context
        self.verbose = context and getattr(context, "verbose", False)
        self.logger = getattr(self.context, "logger", None) if self.context else None

        self._kwargs = kwargs
        self._full_event = kwargs.get("full_event")
        self._input_path = kwargs.get("input_path")
        self._result_path = kwargs.get("result_path")
        self._runnable = False
        name = kwargs.get("name", None)
        if name:
            self.name = name
        else:
            self.name = type(self).__name__

        self._closeables = []

    def _init(self):
        self._termination_received = 0
        self._termination_result = None

    def to_dict(self, fields=None, exclude=None):
        """convert the step object to a python dictionary"""
        fields = fields or getattr(self, "_dict_fields", None)
        if not fields:
            fields = list(inspect.signature(self.__init__).parameters.keys())
        if exclude:
            fields = [field for field in fields if field not in exclude]

        meta_keys = [
            "context",
            "name",
            "input_path",
            "result_path",
            "full_event",
            "kwargs",
        ]
        args = {
            key: getattr(self, key) for key in fields if getattr(self, key, None) is not None and key not in meta_keys
        }
        # add storey kwargs or extra kwargs
        if "kwargs" in fields and (hasattr(self, "kwargs") or hasattr(self, "_kwargs")):
            kwargs = getattr(self, "kwargs", {}) or getattr(self, "_kwargs", {})
            for key, value in kwargs.items():
                if key not in meta_keys:
                    args[key] = value

        mod_name = self.__class__.__module__
        class_path = self.__class__.__qualname__
        if mod_name != "__main__":
            class_path = f"{mod_name}.{class_path}"
        struct = {
            "class_name": class_path,
            "name": self.name or self.__class__.__name__,
            "class_args": args,
        }

        for parameter_name in [
            ("kind", "_STEP_KIND"),
            "input_path",
            "result_path",
            "full_event",
        ]:
            if isinstance(parameter_name, tuple):
                parameter_name, field_name = parameter_name
            else:
                field_name = f"_{parameter_name}"
            if hasattr(self, field_name):
                field_value = getattr(self, field_name)
                if field_value is not None:
                    struct[parameter_name] = field_value
        return struct

    def _to_code(self, taken: Set[str]):
        class_name = type(self).__name__
        base_var_name = ""
        for c in class_name:
            if c.isupper() and base_var_name:
                base_var_name += "_"
            base_var_name += c.lower()
        i = 0
        while True:
            var_name = f"{base_var_name}{i}"
            if var_name not in taken:
                taken.add(var_name)
                break
            i += 1
        taken.add(var_name)
        param_list = []
        for key, value in self._kwargs.items():
            if isinstance(value, str):
                value = f"'{value}'"
            param_list.append(f"{key}={value}")
        param_str = ", ".join(param_list)
        step = f"{var_name} = {class_name}({param_str})"
        steps = [step]
        tos = []
        for outlet in self._outlets:
            outlet_var_name, outlet_steps, outlet_tos = outlet._to_code(taken)
            steps.extend(outlet_steps)
            tos.append(f"{var_name}.to({outlet_var_name})")
            tos.extend(outlet_tos)
        return var_name, steps, tos

    def to_code(self):
        _, steps, tos = self._to_code(set())
        result = "\n".join(steps)
        result += "\n\n"
        result += "\n".join(tos)
        result += "\n"
        return result

    def to(self, outlet):
        if outlet._legal_first_step:
            raise ValueError(f"{outlet.name} can only appear as the first step of a flow")
        self._outlets.append(outlet)
        outlet._inlets.append(self)
        return outlet

    def set_recovery_step(self, outlet):
        self._recovery_step = outlet
        return self

    def _get_recovery_step(self, exception):
        if isinstance(self._recovery_step, dict):
            return self._recovery_step.get(type(exception), None)
        else:
            return self._recovery_step

    def run(self):
        if not self._legal_first_step and not self._runnable:
            raise ValueError("Flow must start with a source")
        self._init()
        outlets = []
        outlets.extend(self._outlets)
        outlets.extend(self._get_recovery_steps())
        for outlet in outlets:
            outlet._runnable = True
            self._closeables.extend(outlet.run())
        return self._closeables

    def _get_recovery_steps(self):
        if self._recovery_step:
            if isinstance(self._recovery_step, dict):
                return list(self._recovery_step.values())
            else:
                return [self._recovery_step]
        return []

    async def run_async(self):
        raise NotImplementedError

    async def _do(self, event):
        raise NotImplementedError

    async def _do_and_recover(self, event):
        try:
            return await self._do(event)
        except BaseException as ex:
            if getattr(ex, "_raised_by_storey_step", None) is not None:
                raise ex
            ex._raised_by_storey_step = self
            recovery_step = self._get_recovery_step(ex)
            if recovery_step is None:
                if self.context and hasattr(self.context, "push_error"):
                    message = traceback.format_exc()
                    if event._awaitable_result:
                        none_or_coroutine = event._awaitable_result._set_error(ex)
                        if none_or_coroutine:
                            await none_or_coroutine
                    if self.logger:
                        self.logger.error(f"Pushing error to error stream: {ex}\n{message}")
                    self.context.push_error(event, f"{ex}\n{message}", source=self.name)
                    return
                else:
                    raise ex
            event.origin_state = self.name
            event.error = ex
            return await recovery_step._do(event)

    @staticmethod
    def _event_string(event):
        result = "Event("
        if event.id:
            result += f"id={event.id}, "
        if getattr(event, "key", None):
            result += f"key={event.key}, "
        if getattr(event, "time", None):
            result += f"processing_time={event.processing_time}, "
        if getattr(event, "path", None):
            result += f"path={event.path}, "
        result += f"body={event.body})"
        return result

    def _should_terminate(self):
        return self._termination_received == len(self._inlets)

    async def _do_downstream(self, event):
        if not self._outlets:
            return
        if event is _termination_obj:
            # Only propagate the termination object once we received one per inlet
            self._outlets[0]._termination_received += 1
            if self._outlets[0]._should_terminate():
                self._termination_result = await self._outlets[0]._do(_termination_obj)
            for outlet in self._outlets[1:] + self._get_recovery_steps():
                outlet._termination_received += 1
                if outlet._should_terminate():
                    self._termination_result = self._termination_result_fn(
                        self._termination_result, await outlet._do(_termination_obj)
                    )
            return self._termination_result
        # If there is more than one outlet, allow concurrent execution.
        tasks = []
        if len(self._outlets) > 1:
            awaitable_result = event._awaitable_result
            event._awaitable_result = None
            original_events = getattr(event, "_original_events", None)
            # Temporarily delete self-reference to avoid deepcopy getting stuck in an infinite loop
            event._original_events = None
            for i in range(1, len(self._outlets)):
                event_copy = copy.deepcopy(event)
                event_copy._awaitable_result = awaitable_result
                event_copy._original_events = original_events
                tasks.append(asyncio.get_running_loop().create_task(self._outlets[i]._do_and_recover(event_copy)))
            # Set self-reference back after deepcopy
            event._original_events = original_events
            event._awaitable_result = awaitable_result
        if self.verbose and self.logger:
            step_name = self.name
            event_string = self._event_string(event)
            self.logger.debug(f"{step_name} -> {self._outlets[0].name} | {event_string}")
        await self._outlets[0]._do_and_recover(event)  # Optimization - avoids creating a task for the first outlet.
        for i, task in enumerate(tasks, start=1):
            if self.verbose and self.logger:
                self.logger.debug(f"{step_name} -> {self._outlets[i].name} | {event_string}")
            await task

    def _get_event_or_body(self, event):
        if self._full_event:
            return event
        elif self._input_path:
            if not hasattr(event.body, "__getitem__"):
                raise TypeError("input_path parameter supports only dict-like event bodies")
            return get_in(event.body, self._input_path)
        else:
            return event.body

    def _user_fn_output_to_event(self, event, fn_result):
        if self._full_event:
            return fn_result
        else:
            mapped_event = copy.copy(event)
            if self._result_path:
                if not hasattr(event.body, "__getitem__"):
                    raise TypeError("result_path parameter supports only dict-like event bodies")
                update_in(mapped_event.body, self._result_path, fn_result)
            else:
                mapped_event.body = fn_result
            return mapped_event

    def _check_step_in_flow(self, type_to_check):
        if isinstance(self, type_to_check):
            return True
        for outlet in self._outlets:
            if outlet._check_step_in_flow(type_to_check):
                return True
        if isinstance(self._recovery_step, Flow):
            if self._recovery_step._check_step_in_flow(type_to_check):
                return True
        elif isinstance(self._recovery_step, dict):
            for step in self._recovery_step.values():
                if step._check_step_in_flow(type_to_check):
                    return True
        return False


class WithUUID:
    def __init__(self):
        self._current_uuid_base = None
        self._current_uuid_count = 0

    def _get_uuid(self):
        if not self._current_uuid_base or self._current_uuid_count == 1024:
            self._current_uuid_base = uuid.uuid4().hex
            self._current_uuid_count = 0
        result = f"{self._current_uuid_base}-{self._current_uuid_count:04}"
        self._current_uuid_count += 1
        return result


class Choice(Flow):
    """Redirects each input element into at most one of multiple downstreams.

    :param choice_array: a list of (downstream, condition) tuples, where downstream is a step and condition is a
        function. The first condition in the list to evaluate as true for an input element causes that element to
        be redirected to that downstream step.
    :type choice_array: tuple of (Flow, Function (Event=>boolean))
    :param default: a default step for events that did not match any condition in choice_array. If not set, elements
        that don't match any condition will be discarded.
    :type default: Flow
    :param name: Name of this step, as it should appear in logs. Defaults to class name (Choice).
    :type name: string
    :param full_event: Whether user functions should receive and return Event objects (when True),
        or only the payload (when False). Defaults to False.
    :type full_event: boolean
    """

    def __init__(self, choice_array, default=None, **kwargs):
        Flow.__init__(self, **kwargs)

        self._choice_array = choice_array
        for outlet, _ in choice_array:
            self.to(outlet)

        if default:
            self.to(default)
        self._default = default

    async def _do(self, event):
        if not self._outlets or event is _termination_obj:
            return await super()._do_downstream(event)
        chosen_outlet = None
        element = self._get_event_or_body(event)
        for outlet, condition in self._choice_array:
            if condition(element):
                chosen_outlet = outlet
                break
        if chosen_outlet:
            await chosen_outlet._do(event)
        elif self._default:
            await self._default._do(event)


class Recover(Flow):
    def __init__(self, exception_to_downstream, **kwargs):
        Flow.__init__(self, **kwargs)

        self._exception_to_downstream = exception_to_downstream

    async def _do(self, event):
        if not self._outlets or event is _termination_obj:
            return await super()._do_downstream(event)
        else:
            try:
                await super()._do_downstream(event)
            except BaseException as ex:
                typ = type(ex)
                if typ in self._exception_to_downstream:
                    await self._exception_to_downstream[typ]._do(event)
                else:
                    raise ex


class _UnaryFunctionFlow(Flow):
    def __init__(self, fn, long_running=None, pass_context=None, **kwargs):
        super().__init__(**kwargs)
        if not callable(fn):
            raise TypeError(f"Expected a callable, got {type(fn)}")
        self._is_async = asyncio.iscoroutinefunction(fn)
        if self._is_async and long_running:
            raise ValueError("long_running=True cannot be used in conjunction with a coroutine")
        self._long_running = long_running
        self._fn = fn
        self._pass_context = pass_context

    async def _call(self, element):
        if self._long_running:
            res = await asyncio.get_running_loop().run_in_executor(None, self._fn, element)
        else:
            kwargs = {}
            if self._pass_context:
                kwargs = {"context": self.context}
            res = self._fn(element, **kwargs)
        if self._is_async:
            res = await res
        return res

    async def _do_internal(self, element, fn_result):
        raise NotImplementedError()

    async def _do(self, event):
        if event is _termination_obj:
            return await self._do_downstream(_termination_obj)
        else:
            element = self._get_event_or_body(event)
            fn_result = await self._call(element)
            await self._do_internal(event, fn_result)


class DropColumns(Flow):
    def __init__(self, columns, **kwargs):
        super().__init__(**kwargs)
        if not isinstance(columns, list):
            columns = [columns]
        self._columns = columns

    async def _do(self, event):
        if event is not _termination_obj:
            new_body = copy.copy(event.body)
            for column in self._columns:
                new_body.pop(column, None)
            event.body = new_body
        return await self._do_downstream(event)


class Map(_UnaryFunctionFlow):
    """Maps, or transforms, incoming events using a user-provided function.

    :param fn: Function to apply to each event
    :type fn: Function (Event=>Event)
    :param long_running: Whether fn is a long-running function. Long-running functions are run in an executor to
        avoid blocking other concurrent processing. Default is False.
    :type long_running: boolean
    :param name: Name of this step, as it should appear in logs. Defaults to class name (Map).
    :type name: string
    :param full_event: Whether user functions should receive and return Event objects (when True),
        or only the payload (when False). Defaults to False.
    :type full_event: boolean
    """

    async def _do_internal(self, event, fn_result):
        mapped_event = self._user_fn_output_to_event(event, fn_result)
        await self._do_downstream(mapped_event)


class Filter(_UnaryFunctionFlow):
    """Filters events based on a user-provided function.

    :param fn: Function to decide whether to keep each event.
    :type fn: Function (Event=>boolean)
    :param long_running: Whether fn is a long-running function. Long-running functions are run in an executor to
        avoid blocking other concurrent processing. Default is False.
    :type long_running: boolean
    :param name: Name of this step, as it should appear in logs. Defaults to class name (Filter).
    :type name: string
    :param full_event: Whether user functions should receive and return Event objects (when True), or only the
        payload (when False). Defaults to False.
    :type full_event: boolean
    """

    async def _do_internal(self, event, keep):
        if keep:
            await self._do_downstream(event)


class FlatMap(_UnaryFunctionFlow):
    """Maps, or transforms, each incoming event into any number of events.

    :param fn: Function to transform each event to a list of events.
    :type fn: Function (Event=>list of Event)
    :param long_running: Whether fn is a long-running function. Long-running functions are run in an executor
        to avoid blocking other concurrent processing. Default is False.
    :type long_running: boolean
    :param name: Name of this step, as it should appear in logs. Defaults to class name (FlatMap).
    :type name: string
    :param full_event: Whether user functions should receive and return Event objects (when True), or only
        the payload (when False). Defaults to False.
    :type full_event: boolean
    """

    async def _do_internal(self, event, fn_result):
        for fn_result_element in fn_result:
            mapped_event = self._user_fn_output_to_event(event, fn_result_element)
            await self._do_downstream(mapped_event)


class Extend(_UnaryFunctionFlow):
    """Adds fields to each incoming event.

    :param fn: Function to transform each event to a dictionary. The fields in the returned dictionary are then added
        to the original event.
    :type fn: Function (Event=>Dict)
    :param long_running: Whether fn is a long-running function. Long-running functions are run in an executor to avoid
        blocking other concurrent processing. Default is False.
    :type long_running: boolean
    :param name: Name of this step, as it should appear in logs. Defaults to class name (Extend).
    :type name: string
    :param full_event: Whether user functions should receive and return Event objects (when True), or only the
        payload (when False). Defaults to False.
    :type full_event: boolean
    """

    async def _do_internal(self, event, fn_result):
        for key, value in fn_result.items():
            event.body[key] = value
        await self._do_downstream(event)


class _FunctionWithStateFlow(Flow):
    def __init__(self, initial_state, fn, group_by_key=False, **kwargs):
        super().__init__(**kwargs)
        if not callable(fn):
            raise TypeError(f"Expected a callable, got {type(fn)}")
        self._is_async = asyncio.iscoroutinefunction(fn)
        self._state = initial_state
        if isinstance(self._state, str):
            should_get_from_context = False
            for known_scheme in known_driver_schemes:
                if self._state.startswith(f"{known_scheme}://"):
                    should_get_from_context = True
                    break
            if should_get_from_context:
                if not self.context:
                    raise TypeError("Table can not be string if no context was provided to the step")
                self._state = self.context.get_table(self._state)
        self._fn = fn
        self._group_by_key = group_by_key
        if hasattr(self._state, "close"):
            self._closeables = [self._state]

    async def _call(self, event):
        element = self._get_event_or_body(event)
        if self._group_by_key:
            safe_key = stringify_key(event.key)
            if isinstance(self._state, Table):
                key_data = await self._state._get_or_load_static_attributes_by_key(safe_key)
            else:
                key_data = self._state[event.key]
            res, new_state = self._fn(element, key_data)
            async with self._state._get_lock(safe_key):
                self._state._update_static_attrs(safe_key, new_state)
                self._state._pending_events.append(event)
            self._state._init_flush_task()
        else:
            res, self._state = self._fn(element, self._state)
        if self._is_async:
            res = await res
        return res

    async def _do_internal(self, element, fn_result):
        raise NotImplementedError()

    async def _do(self, event):
        if event is _termination_obj:
            return await self._do_downstream(_termination_obj)
        else:
            fn_result = await self._call(event)
            await self._do_internal(event, fn_result)


class MapWithState(_FunctionWithStateFlow):
    """Maps, or transforms, incoming events using a stateful user-provided function, and an initial state,
        which may be a database table.

    :param initial_state: Initial state for the computation. If group_by_key is True, this must be a dictionary or
        a Table object.
    :type initial_state: dictionary or Table if group_by_key is True. Any object otherwise.
    :param fn: A function to run on each event and the current state. Must yield an event and an updated state.
    :type fn: Function ((Event, state)=>(Event, state))
    :param group_by_key: Whether the state is computed by key. Optional. Default to False.
    :type group_by_key: boolean
    :param full_event: Whether fn will receive and return an Event object or only the body (payload).
        Optional. Defaults to False (body only).
    :type full_event: boolean
    """

    async def _do_internal(self, event, mapped_element):
        mapped_event = self._user_fn_output_to_event(event, mapped_element)
        await self._do_downstream(mapped_event)


class MapClass(Flow):
    """Similar to Map, but instead of a function argument, this class should be extended and its do()
    method overridden."""

    def __init__(self, long_running=None, **kwargs):
        super().__init__(**kwargs)
        self._is_async = asyncio.iscoroutinefunction(self.do)
        if self._is_async and long_running:
            raise ValueError("long_running=True cannot be used in conjunction with a coroutine do()")
        self._long_running = long_running
        self._filter = False

    def filter(self):
        # used in the .do() code to signal filtering
        self._filter = True

    def do(self, event):
        raise NotImplementedError()

    async def _call(self, event):
        if self._long_running:
            res = await asyncio.get_running_loop().run_in_executor(None, self.do, event)
        else:
            res = self.do(event)
        if self._is_async:
            res = await res
        return res

    async def _do(self, event):
        if event is _termination_obj:
            return await self._do_downstream(_termination_obj)
        else:
            element = self._get_event_or_body(event)
            fn_result = await self._call(element)
            if not self._filter:
                mapped_event = self._user_fn_output_to_event(event, fn_result)
                await self._do_downstream(mapped_event)
            else:
                self._filter = False  # clear the flag for future runs


class Rename(Flow):
    """
    Rename fields in event body.

    :param mapping: Dictionary from old name to new name.
    :param name: Name of this step, as it should appear in logs. Defaults to class name (Rename).
    :type name: string
    """

    def __init__(self, mapping: Dict[str, str], **kwargs):
        super().__init__(**kwargs)
        self.mapping = mapping

    async def _do(self, event):
        if event is not _termination_obj:
            for old_name, new_name in self.mapping.items():
                if old_name in event.body:
                    event.body[new_name] = event.body.get(old_name)
                    del event.body[old_name]
        return await self._do_downstream(event)


class ReifyMetadata(Flow):
    """
    Inserts event metadata into the event body.

    :param mapping: Dictionary from event attribute name to entry key in the event body (which must be a
        dictionary). Alternatively, an iterable of names may be provided, and these will be used as both
        attribute name and entry key.
    :param name: Name of this step, as it should appear in logs. Defaults to class name (ReifyMetadata).
    :type name: string
    """

    def __init__(self, mapping: Iterable[str], **kwargs):
        super().__init__(**kwargs)
        self.mapping = mapping

    async def _do(self, event):
        if event is not _termination_obj:
            if isinstance(self.mapping, dict):
                for attribute_name, entry_key in self.mapping.items():
                    event.body[entry_key] = getattr(event, attribute_name)
            else:
                for attribute_name in self.mapping:
                    event.body[attribute_name] = getattr(event, attribute_name)
        return await self._do_downstream(event)


class Complete(Flow):
    """
    Completes the AwaitableResult associated with incoming events.

    :param name: Name of this step, as it should appear in logs. Defaults to class name (Complete).
    :type name: string
    :param full_event: Whether to complete with an Event object (when True) or only the payload
        (when False). Default to False.
    :type full_event: boolean
    """

    async def _do(self, event):
        termination_result = await self._do_downstream(event)
        if event is not _termination_obj:
            if event._awaitable_result:
                result = self._get_event_or_body(event)
                res = event._awaitable_result._set_result(result)
                if res:
                    await res
        return termination_result


class Reduce(Flow):
    """
    Reduces incoming events into a single value which is returned upon the successful termination of the flow.

    :param initial_value: Starting value. When the first event is received, fn will be applied to the
        initial_value and that event.
    :type initial_value: object
    :param fn: Function to apply to the current value and each event.
    :type fn: Function ((object, Event) => object)
    :param name: Name of this step, as it should appear in logs. Defaults to class name (Reduce).
    :type name: string
    :param full_event: Whether user functions should receive and return Event objects (when True),
        or only the payload (when False). Defaults to False.
    :type full_event: boolean
    """

    def __init__(self, initial_value, fn, **kwargs):
        kwargs["initial_value"] = initial_value
        super().__init__(**kwargs)
        if not callable(fn):
            raise TypeError(f"Expected a callable, got {type(fn)}")
        self._is_async = asyncio.iscoroutinefunction(fn)
        self._fn = fn
        self._initial_value = initial_value

    def _init(self):
        super()._init()
        self._result = self._initial_value

    def to(self, outlet):
        raise ValueError("Reduce is a terminal step. It cannot be piped further.")

    async def _do(self, event):
        if event is _termination_obj:
            return self._result
        else:
            if self._full_event:
                elem = event
            else:
                elem = event.body
            res = self._fn(self._result, elem)
            if self._is_async:
                res = await res
            self._result = res


class HttpRequest:
    """A class representing an HTTP request, with method, url, body, and headers.

    :param method: HTTP method (e.g. GET).
    :type method: string
    :param url: Target URL (http and https schemes supported).
    :type url: string
    :param body: Request body.
    :type body: bytes or string
    :param headers: Request headers, in the form of a dictionary. Optional. Defaults to no headers.
    :type headers: dictionary, or None.
    """

    def __init__(self, method, url, body, headers: Optional[dict] = None):
        self.method = method
        self.url = url
        self.body = body
        if headers is None:
            headers = {}
        self.headers = headers


class HttpResponse:
    """A class representing an HTTP response, with a status code and body.

    :param body: Response body.
    :type body: bytes
    :param status: HTTP status code.
    :type status: int
    """

    def __init__(self, status, body):
        self.status = status
        self.body = body


class _ConcurrentJobExecution(Flow):
    _BACKOFF_MAX = 120
    _DEFAULT_MAX_IN_FLIGHT = 8

    def __init__(self, max_in_flight=None, retries=None, backoff_factor=None, **kwargs):
        Flow.__init__(self, **kwargs)
        if max_in_flight is not None and max_in_flight < 1:
            raise ValueError(f"max_in_flight may not be less than 1 (got {max_in_flight})")
        self.max_in_flight = max_in_flight
        self.retries = retries
        self.backoff_factor = backoff_factor

        self._queue_size = (max_in_flight or self._DEFAULT_MAX_IN_FLIGHT) - 1

    def _init(self):
        super()._init()
        self._q = None
        self._lazy_init_complete = False

    async def _worker(self):
        event = None
        try:
            while True:
                try:
                    # If we don't handle the event before we remove it from the queue, the effective max_in_flight will
                    # be 1 higher than requested. Hence, we peek.
                    job = await self._q.peek()
                    if job is _termination_obj:
                        await self._q.get()
                        break
                    event = job[0]
                    completed = await job[1]
                    await self._handle_completed(event, completed)
                    await self._q.get()
                except BaseException as ex:
                    await self._q.get()
                    ex._raised_by_storey_step = self
                    recovery_step = self._get_recovery_step(ex)
                    try:
                        if recovery_step is not None:
                            event.origin_state = self.name
                            event.error = ex
                            await recovery_step._do(event)
                        else:
                            if event._awaitable_result:
                                none_or_coroutine = event._awaitable_result._set_error(ex)
                                if none_or_coroutine:
                                    await none_or_coroutine
                            if self.context and hasattr(self.context, "push_error"):
                                message = traceback.format_exc()
                                if self.logger:
                                    self.logger.error(f"Pushing error to error stream: {ex}\n{message}")
                                self.context.push_error(event, f"{ex}\n{message}", source=self.name)
                            else:
                                raise ex
                    except BaseException:
                        if not self._q.empty():
                            await self._q.get()
                        raise
        finally:
            await self._cleanup()

    async def _process_event(self, event):
        raise NotImplementedError()

    async def _handle_completed(self, event, response):
        raise NotImplementedError()

    async def _cleanup(self):
        pass

    async def _lazy_init(self):
        pass

    async def _process_event_with_retries(self, event):
        times_attempted = 0
        max_attempts = (self.retries or 0) + 1
        while True:
            try:
                return await self._process_event(event)
            except Exception as ex:
                times_attempted += 1
                attempts_left = max_attempts - times_attempted
                if self.logger:
                    self.logger.warn(f"{self.name} failed to process event ({attempts_left} retries left): {ex}")
                if attempts_left <= 0:
                    raise ex
                backoff_value = (self.backoff_factor or 1) * (2 ** (times_attempted - 1))
                backoff_value = min(self._BACKOFF_MAX, backoff_value)
                if backoff_value >= 0:
                    await asyncio.sleep(backoff_value)

    async def _do(self, event):
        if not self._lazy_init_complete:
            await self._lazy_init()
            self._lazy_init_complete = True

        if not self._q and self._queue_size > 0:
            self._q = AsyncQueue(self._queue_size)
            self._worker_awaitable = asyncio.get_running_loop().create_task(self._worker())

        if self._queue_size > 0 and self._worker_awaitable.done():
            await self._worker_awaitable
            raise FlowError("ConcurrentJobExecution worker has already terminated")

        if event is _termination_obj:
            if self._queue_size > 0:
                await self._q.put(_termination_obj)
                await self._worker_awaitable
            else:
                await self._cleanup()
            return await self._do_downstream(_termination_obj)
        else:
            coroutine = self._process_event_with_retries(event)
            if self._queue_size == 0:
                completed = await coroutine
                await self._handle_completed(event, completed)
            else:
                task = asyncio.get_running_loop().create_task(coroutine)
                await self._q.put((event, task))
                if self._worker_awaitable.done():
                    await self._worker_awaitable


class ConcurrentExecution(_ConcurrentJobExecution):
    """
    Inherit this class and override `process_event()` to process events concurrently.

    :param process_event: Function that will be run on each event

    :param concurrency_mechanism: One of:
      * "asyncio" (default) – for I/O implemented using asyncio
      * "threading" – for blocking I/O
      * "multiprocessing" – for processing-intensive tasks

    :param max_in_flight: Maximum number of events to be processed at a time (default 8)
    :param retries: Maximum number of retries per event (default 0)
    :param backoff_factor: Wait time in seconds before the first retry (default 1). Subsequent retries will each wait
      twice as the previous retry, up to a maximum of two minutes.
    :param pass_context: If False, the process_event function will be called with just one parameter (event). If True,
      the process_event function will be called with two parameters (event, context). Defaults to False.
    :param full_event: Whether event processor should receive and return Event objects (when True),
        or only the payload (when False). Defaults to False.
    """

    _supported_concurrency_mechanisms = ["asyncio", "threading", "multiprocessing"]

    def __init__(
        self,
        event_processor: Union[Callable[[Event], Any], Callable[[Event, Any], Any]],
        concurrency_mechanism=None,
        pass_context=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if concurrency_mechanism == "multiprocessing" and kwargs.get("full_event"):
            raise ValueError(
                'concurrency_mechanism="multiprocessing" may not be used in conjunction with full_event=True'
            )

        self._event_processor = event_processor

        if concurrency_mechanism and concurrency_mechanism not in self._supported_concurrency_mechanisms:
            raise ValueError(f"Concurrency mechanism '{concurrency_mechanism}' is not supported")

        if concurrency_mechanism == "multiprocessing" and pass_context:
            try:
                pickle.dumps(self.context)
            except Exception as ex:
                raise ValueError(
                    'When concurrency_mechanism="multiprocessing" is used in conjunction with '
                    "pass_context=True, context must be serializable"
                ) from ex

        self._executor = None
        if concurrency_mechanism == "threading":
            self._executor = ThreadPoolExecutor(max_workers=self.max_in_flight or self._DEFAULT_MAX_IN_FLIGHT)
        elif concurrency_mechanism == "multiprocessing":
            self._executor = ProcessPoolExecutor(max_workers=self.max_in_flight or self._DEFAULT_MAX_IN_FLIGHT)

        self._pass_context = pass_context

    async def _process_event(self, event):
        args = [event if self._full_event else event.body]

        if self._pass_context:
            args.append(self.context)
        if self._executor:
            result = await asyncio.get_running_loop().run_in_executor(self._executor, self._event_processor, *args)
        else:
            result = self._event_processor(*args)

        if asyncio.iscoroutine(result):
            result = await result

        if self._full_event:
            return result
        else:
            event.body = result
            return event

    async def _handle_completed(self, event, response):
        await self._do_downstream(response)


class SendToHttp(_ConcurrentJobExecution):
    """Joins each event with data from any HTTP source. Used for event augmentation.

    :param request_builder: Creates an HTTP request from the event. This request is then sent to its destination.
    :type request_builder: Function (Event=>HttpRequest)
    :param join_from_response: Joins the original event with the HTTP response into a new event.
    :type join_from_response: Function ((Event, HttpResponse)=>Event)
    :param name: Name of this step, as it should appear in logs. Defaults to class name (SendToHttp).
    :type name: string
    :param full_event: Whether user functions should receive and return Event objects (when True),
        or only the payload (when False). Defaults to False.
    :type full_event: boolean
    """

    def __init__(self, request_builder, join_from_response, **kwargs):
        super().__init__(**kwargs)
        self._request_builder = request_builder
        self._join_from_response = join_from_response

    def _init(self):
        super()._init()
        self._client_session = None

    async def _lazy_init(self):
        connector = aiohttp.TCPConnector()
        self._client_session = aiohttp.ClientSession(connector=connector)

    async def _cleanup(self):
        await self._client_session.close()

    async def _process_event(self, event):
        req = self._request_builder(event)
        return await self._client_session.request(req.method, req.url, headers=req.headers, data=req.body, ssl=False)

    async def _handle_completed(self, event, response):
        response_body = await response.text()
        joined_element = self._join_from_response(event.body, HttpResponse(response.status, response_body))
        if joined_element is not None:
            new_event = self._user_fn_output_to_event(event, joined_element)
            await self._do_downstream(new_event)


class _Batching(Flow):
    _do_downstream_per_event = True

    def __init__(
        self,
        max_events: Optional[int] = None,
        flush_after_seconds: Union[int, float, None] = None,
        key_field: Optional[Union[str, Callable[[Event], str]]] = None,
        drop_key_field=False,
        **kwargs,
    ):
        if max_events:
            kwargs["max_events"] = max_events
        if flush_after_seconds is not None:
            kwargs["flush_after_seconds"] = flush_after_seconds
        if isinstance(key_field, str):
            kwargs["key_field"] = key_field
        super().__init__(**kwargs)

        self._max_events = max_events
        self._flush_after_seconds = flush_after_seconds

        if self._flush_after_seconds is not None and self._flush_after_seconds < 0:
            raise ValueError("flush_after_seconds cannot be negative")

        self._extract_key: Optional[Callable[[Event], str]] = self._create_key_extractor(key_field, drop_key_field)

    def _init(self):
        super()._init()
        self._batch: Dict[Optional[str], List[Any]] = defaultdict(list)
        # Keep the original events that make up each batch
        self._batch_events: Dict[Optional[str], List[Any]] = defaultdict(list)
        self._batch_first_event_time: Dict[Optional[str], datetime.datetime] = {}
        self._batch_last_event_time: Dict[Optional[str], datetime.datetime] = {}
        self._batch_start_time: Dict[Optional[str], float] = {}
        self._timeout_task: Optional[Task] = None

    @staticmethod
    def _create_key_extractor(key_field, drop_key_field) -> Callable:
        if key_field is None:
            return lambda event: None
        elif callable(key_field):
            return key_field
        elif isinstance(key_field, str):
            if key_field.startswith("$"):
                attribute = key_field[1:]
                return lambda event: getattr(event, attribute)
            elif drop_key_field:
                return lambda event: event.body.pop(key_field)
            else:
                return lambda event: event.body[key_field]
        else:
            raise ValueError(f"Unsupported key_field type {type(key_field)}")

    async def _emit(self, batch, batch_key, batch_time, batch_events, last_event_time=None):
        raise NotImplementedError

    async def _terminate(self):
        pass

    async def _do(self, event):
        if event is _termination_obj:
            await self._emit_all()
            await self._terminate()
            return await self._do_downstream(_termination_obj)

        key = self._extract_key(event)

        if hasattr(self, "_get_event_time"):
            event_time = self._get_event_time(event)
        else:
            event_time = event.processing_time

        if len(self._batch[key]) == 0:
            self._batch_first_event_time[key] = event_time
            self._batch_start_time[key] = time.monotonic()
            self._batch_last_event_time[key] = event_time
        elif self._batch_last_event_time[key] < event_time:
            self._batch_last_event_time[key] = event_time

        if self._flush_after_seconds is not None and self._timeout_task is None:
            self._timeout_task = asyncio.get_running_loop().create_task(self._sleep_and_emit())

        self._batch[key].append(self._event_to_batch_entry(event))
        self._batch_events[key].append(event)

        if len(self._batch[key]) == self._max_events:
            await self._emit_batch(key)

        if self._do_downstream_per_event:
            await self._do_downstream(event)

    async def _sleep_and_emit(self):
        try:
            while self._batch:
                key = next(iter(self._batch.keys()))
                delta_seconds = time.monotonic() - self._batch_start_time[key]
                if delta_seconds < self._flush_after_seconds:
                    await asyncio.sleep(self._flush_after_seconds - delta_seconds)
                await self._emit_batch(key)
        except Exception:
            message = traceback.format_exc()
            if self.logger:
                self.logger.error(f"Failed to flush batch in step '{self.name}':\n{message}")

        self._timeout_task = None

    def _event_to_batch_entry(self, event):
        return self._get_event_or_body(event)

    async def _emit_batch(self, batch_key: Optional[str] = None):
        batch_to_emit = self._batch.pop(batch_key, None)
        if batch_to_emit is None:
            return
        batch_time = self._batch_first_event_time.pop(batch_key)
        last_event_time = self._batch_last_event_time.pop(batch_key)
        del self._batch_start_time[batch_key]
        await self._emit(batch_to_emit, batch_key, batch_time, self._batch_events[batch_key], last_event_time)
        del self._batch_events[batch_key]

    async def _emit_all(self):
        for key in list(self._batch.keys()):
            await self._emit_batch(key)


class Batch(_Batching, WithUUID):
    """Batches events into lists of up to max_events events. Each emitted list contained max_events events, unless
    flush_after_seconds seconds have passed since the first event in the batch was received, at which the batch is
    emitted with potentially fewer than max_events event.

    :param max_events: Maximum number of events per emitted batch. Set to None to emit all events in one batch on flow
        termination.
    :param flush_after_seconds: Maximum number of seconds to wait before a batch is emitted.
    :param key: The key by which events are grouped. By default (None), events are not grouped.
        Other options may be:
        Set to '$x' to group events by the x attribute of the event. E.g. "$key" or "$path".
        set to other string 'str' to group events by Event.body[str].
        set a Callable[Any, Any] to group events by a a custom key extractor.
    """

    _do_downstream_per_event = False

    def __init__(self, *args, **kwargs):
        _Batching.__init__(self, *args, **kwargs)
        WithUUID.__init__(self)

    async def _emit(self, batch, batch_key, batch_time, batch_events, last_event_time=None):
        event = Event(batch, id=self._get_uuid())
        if not self._full_event:
            # Preserve reference to the original events to avoid early commit of offsets
            event._original_events = batch_events
        return await self._do_downstream(event)


class JoinWithV3IOTable(_ConcurrentJobExecution):
    """Joins each event with a V3IO table. Used for event augmentation.

    :param storage: Database driver.
    :type storage: Driver
    :param key_extractor: Function for extracting the key for table access from an event.
    :type key_extractor: Function (Event=>string)
    :param join_function: Joins the original event with relevant data received from V3IO.
    :type join_function: Function ((Event, dict)=>Event)
    :param table_path: Path to the table in V3IO.
    :type table_path: string
    :param attributes: A comma-separated list of attributes to be requested from V3IO.
        Defaults to '*' (all user attributes).
    :type attributes: string
    :param name: Name of this step, as it should appear in logs. Defaults to class name (JoinWithV3IOTable).
    :type name: string
    :param full_event: Whether user functions should receive and return Event objects (when True), or only
        the payload (when False). Defaults to False.
    :type full_event: boolean
    """

    def __init__(
        self,
        storage,
        key_extractor,
        join_function,
        table_path,
        attributes="*",
        **kwargs,
    ):
        kwargs["table_path"] = table_path
        kwargs["attributes"] = attributes
        super().__init__(**kwargs)

        self._storage = storage

        self._key_extractor = key_extractor
        self._join_function = join_function

        self._container, self._table_path = _split_path(table_path)
        self._attributes = attributes

    async def _process_event(self, event):
        key = str(self._key_extractor(self._get_event_or_body(event)))
        return await self._storage._get_item(self._container, self._table_path, key, self._attributes)

    async def _handle_completed(self, event, response):
        if response.status_code == 200:
            response_object = response.output.item
            joined = self._join_function(self._get_event_or_body(event), response_object)
            if joined is not None:
                new_event = self._user_fn_output_to_event(event, joined)
                await self._do_downstream(new_event)
        elif response.status_code == 404:
            return None
        else:
            raise V3ioError(f"Failed to get item. Response status code was {response.status_code}: {response.body}")

    async def _cleanup(self):
        await self._storage.close()


class JoinWithTable(_ConcurrentJobExecution):
    """Joins each event with data from the given table.

    :param table: A Table object or name to join with. If a table name is provided, it will be looked up in the context.
    :param key_extractor: Key's column name or a function for extracting the key, for table access from an event.
    :param attributes: A comma-separated list of attributes to be queried for. Defaults to all attributes.
    :param inner_join: Whether to drop events when the table does not have a matching entry (join_function won't be
        called in such a case). Defaults to False.
    :param join_function: Joins the original event with relevant data received from the storage. Event is dropped when
        this function returns None. Defaults to assume the event's body is a dict-like object and updating it.
    :param name: Name of this step, as it should appear in logs. Defaults to class name (JoinWithTable).
    :param full_event: Whether user functions should receive and return Event objects (when True), or only the
        payload (when False). Defaults to False.
    :param context: Context object that holds global configurations and secrets.
    """

    def __init__(
        self,
        table: Union[Table, str],
        key_extractor: Union[str, Callable[[Event], str]],
        attributes: Optional[List[str]] = None,
        inner_join: bool = False,
        join_function: Optional[Callable[[Any, Dict[str, object]], Any]] = None,
        **kwargs,
    ):
        if isinstance(table, str):
            kwargs["table"] = table
        if isinstance(key_extractor, str):
            kwargs["key_extractor"] = key_extractor
        if attributes:
            kwargs["attributes"] = attributes
        kwargs["inner_join"] = inner_join

        super().__init__(**kwargs)

        self._table = table
        if isinstance(table, str):
            if not self.context:
                raise TypeError("Table can not be string if no context was provided to the step")
            self._table = self.context.get_table(table)

        if key_extractor:
            if callable(key_extractor):
                self._key_extractor = key_extractor
            elif isinstance(key_extractor, str):
                if self._full_event:
                    self._key_extractor = lambda event: event.body[key_extractor]
                else:
                    self._key_extractor = lambda element: element[key_extractor]
            else:
                raise TypeError(f"key is expected to be either a callable or string but got {type(key_extractor)}")

        def default_join_fn(event, join_res):
            event.update(join_res)
            return event

        def default_join_fn_full_event(event, join_res):
            event.body.update(join_res)
            return event

        self._inner_join = inner_join
        self._join_function = join_function or (default_join_fn_full_event if self._full_event else default_join_fn)

        self._attributes = attributes or "*"

    def _init(self):
        super()._init()
        self._closeables = [self._table]

    async def _process_event(self, event):
        key = self._key_extractor(self._get_event_or_body(event))
        safe_key = stringify_key(key)
        return await self._table._get_or_load_static_attributes_by_key(safe_key, self._attributes)

    async def _handle_completed(self, event, response):
        if self._inner_join and not response:
            return
        joined = self._join_function(self._get_event_or_body(event), response)
        if joined is not None:
            new_event = self._user_fn_output_to_event(event, joined)
            await self._do_downstream(new_event)


def build_flow(steps):
    """Builds a flow from a list of steps, by chaining the steps according to their order in the list.
    Nested lists are used to represent branches in the flow.

    Examples:
        build_flow([step1, step2, step3])
        is equivalent to
        step1.to(step2).to(step3)

        build_flow([step1, [step2a, step2b], step3])
        is equivalent to
        step1.to(step2a)
        step1.to(step3)
        step2a.to(step2b)

    :param steps: a potentially nested list of steps
    :returns: the first step
    :rtype: Flow
    """
    if len(steps) == 0:
        raise ValueError("Cannot build an empty flow")
    first_step = steps[0]
    if isinstance(first_step, list):
        first_step = build_flow(first_step)
    cur_step = first_step
    for next_step in steps[1:]:
        if isinstance(next_step, list):
            cur_step.to(build_flow(next_step))
        else:
            cur_step.to(next_step)
            cur_step = next_step
    return first_step


class Context:
    """
    Context object that holds global secrets and configurations to be passed to relevant steps.

    :param initial_secrets: Initial dict of secrets.
    :param initial_parameters: Initial dict of parameters.
    :param initial_tables: Initial dict of tables.
    """

    def __init__(
        self,
        initial_secrets: Optional[Dict[str, str]] = None,
        initial_parameters: Optional[Dict[str, object]] = None,
        initial_tables: Optional[Dict[str, Table]] = None,
    ):
        self._secrets = initial_secrets or {}
        self._parameters = initial_parameters or {}
        self._tables = initial_tables or {}

    def get_param(self, key, default):
        return self._parameters.get(key, default)

    def set_param(self, key, value):
        self._parameters[key] = value

    def get_secret(self, key):
        return self._secrets.get(key, None)

    def set_secret(self, key, secret):
        self._secrets[key] = secret

    def get_table(self, key):
        return self._tables[key]

    def set_table(self, key, table):
        self._tables[key] = table
